from enum import Enum
import enum
from pathlib import Path
from types import NoneType
import numpy as np
from pydantic import BaseModel, Field
from typing import Any, Literal, Optional, Type, Union

from nodetool.models.asset import Asset
from nodetool.models.message import Message as MessageModel, MessageContent
from nodetool.models.task import Task as TaskModel


# Mapping of python types to their string representation
TypeToName = {}
NameToType = {}


def add_type_name(type: Type, name: str):
    """
    Adds a type name to the TypeToEnum and EnumToType mappings.
    """
    TypeToName[type] = name
    NameToType[name] = type


def add_type_names(types):
    """
    Add type names to the TypeToEnum and EnumToType mappings.
    """
    for type, name in types.items():
        add_type_name(type, name)


# Add the default type names
add_type_names(
    {
        Any: "any",
        NoneType: "none",
        list: "list",
        dict: "dict",
        int: "int",
        float: "float",
        bool: "bool",
        str: "str",
        Enum: "enum",
        Union: "union",
    }
)


class BaseType(BaseModel):
    """
    This is the base class for all Nodetool types.

    It is used to create a mapping of type names to their corresponding classes.
    """

    type: str

    @classmethod
    def __init_subclass__(cls):
        """
        This method is called when a subclass of BaseType is created.
        We remember the mapping of the subclass to its type name,
        so that we can use it later to create instances of the subclass from the type name.
        """
        super().__init_subclass__()
        if hasattr(cls, "type"):
            add_type_name(cls, cls.type)


asset_types = set()


class AssetRef(BaseType):
    type: str = "asset"
    uri: str = ""
    asset_id: str | None = None
    temp_id: str | None = None

    def to_dict(self):
        res = {
            "uri": self.uri,
        }
        if self.asset_id:
            res["asset_id"] = self.asset_id
        if self.temp_id:
            res["temp_id"] = self.temp_id
        return res

    def is_empty(self):
        return self.uri == "" and self.asset_id is None

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        if hasattr(cls, "type"):
            asset_types.add(cls.type)


class FolderRef(AssetRef):
    type: Literal["folder"] = "folder"


class ModelRef(AssetRef):
    type: Literal["model_ref"] = "model_ref"


class VideoRef(AssetRef):
    type: Literal["video"] = "video"
    duration: Optional[float] = None  # Duration in seconds
    format: Optional[str] = None


class TextRef(AssetRef):
    type: Literal["text"] = "text"


class AudioRef(AssetRef):
    type: Literal["audio"] = "audio"


class ImageRef(AssetRef):
    """A reference to an image asset."""

    type: Literal["image"] = "image"


class WorkflowRef(BaseType):
    type: Literal["workflow"] = "workflow"
    id: str = ""


class NodeRef(BaseType):
    type: Literal["node"] = "node"
    id: str = ""


class FileRef(BaseType):
    type: Literal["file"] = "file"


class ModelFile(BaseType):
    name: str = ""


class GPTModel(str, enum.Enum):
    GPT3 = "gpt-3.5-turbo-0125"
    GPT4 = "gpt-4o"


class LanguageModel(ModelFile):
    type: Literal["language_model"] = "language_model"


class FunctionModel(LanguageModel):
    type: Literal["function_model"] = "function_model"
    repo_id: str = ""
    filename: str = ""
    local_path: Path | None = None


class LlamaModel(LanguageModel):
    type: Literal["llama_model"] = "llama_model"
    repo_id: str = ""
    filename: str = ""
    local_path: Path | None = None


class CheckpointFile(ModelFile):
    type: Literal["comfy.checkpoint_file"] = "comfy.checkpoint_file"


class VAEFile(ModelFile):
    type: Literal["comfy.vae_file"] = "comfy.vae_file"


class CLIPFile(ModelFile):
    type: Literal["comfy.clip_file"] = "comfy.clip_file"


class unCLIPFile(ModelFile):
    type: Literal["comfy.unclip_file"] = "comfy.unclip_file"


class GLIGENFile(ModelFile):
    type: Literal["comfy.gligen_file"] = "comfy.gligen_file"


class CLIPVisionFile(ModelFile):
    type: Literal["comfy.clip_vision_file"] = "comfy.clip_vision_file"


class ControlNetFile(ModelFile):
    type: Literal["comfy.control_net_file"] = "comfy.control_net_file"


class IPAdapterFile(ModelFile):
    type: Literal["comfy.ip_adapter_file"] = "comfy.ip_adapter_file"


class LORAFile(ModelFile):
    type: Literal["comfy.lora_file"] = "comfy.lora_file"


class UpscaleModelFile(ModelFile):
    type: Literal["comfy.upscale_model_file"] = "comfy.upscale_model_file"


class CLIP(BaseType):
    type: Literal["comfy.clip"] = "comfy.clip"


class CLIPVision(BaseType):
    type: Literal["comfy.clip_vision"] = "comfy.clip_vision"


class CLIPVisionOutput(BaseType):
    type: Literal["comfy.clip_vision_output"] = "comfy.clip_vision_output"


class GLIGEN(BaseType):
    type: Literal["comfy.gligen"] = "comfy.gligen"


class Conditioning(BaseType):
    type: Literal["comfy.conditioning"] = "comfy.conditioning"


class Latent(BaseType):
    type: Literal["comfy.latent"] = "comfy.latent"


class ControlNet(BaseType):
    type: Literal["comfy.control_net"] = "comfy.control_net"


class VAE(BaseType):
    type: Literal["comfy.vae"] = "comfy.vae"


class UNet(BaseType):
    type: Literal["comfy.unet"] = "comfy.unet"


class ImageTensor(BaseType):
    type: Literal["comfy.image_tensor"] = "comfy.image_tensor"


class UpscaleModel(BaseType):
    type: Literal["comfy.upscale_model"] = "comfy.upscale_model"


class LORA(BaseType):
    type: Literal["comfy.lora"] = "comfy.lora"


class IPAdapter(BaseType):
    type: Literal["comfy.ip_adapter"] = "comfy.ip_adapter"


class InsightFace(BaseType):
    type: Literal["comfy.insight_face"] = "comfy.insight_face"


class Mask(BaseType):
    type: Literal["comfy.mask"] = "comfy.mask"


class Sigmas(BaseType):
    type: Literal["comfy.sigmas"] = "comfy.sigmas"


class Sampler(BaseType):
    type: Literal["comfy.sampler"] = "comfy.sampler"


class Embeds(BaseType):
    type: Literal["comfy.embeds"] = "comfy.embeds"


class Task(BaseType):
    type: Literal["task"] = "task"
    id: str = ""
    task_type: str = ""
    user_id: str = ""
    thread_id: str = ""
    status: str = ""
    name: str = ""
    instructions: str = ""
    dependencies: list[str] = []
    required_capabilities: list[str] = []
    started_at: str = ""
    finished_at: str | None = None
    error: str | None = None
    result: str | None = None
    cost: float | None = None

    @classmethod
    def from_model(cls, task: TaskModel):
        return cls(
            id=task.id,
            user_id=task.user_id,
            task_type=task.task_type,
            thread_id=task.thread_id,
            status=task.status,
            name=task.name,
            instructions=task.instructions,
            dependencies=task.dependencies,
            required_capabilities=task.required_capabilities,
            started_at=task.started_at.isoformat(),
            finished_at=task.finished_at.isoformat() if task.finished_at else None,
            error=task.error,
            result=task.result,
            cost=task.cost,
        )


class ToolCall(BaseType):
    type: Literal["tool_call"] = "tool_call"
    function_name: str = ""
    function_args: dict[str, Any] = {}
    function_response: Any = None


class Tensor(BaseType):
    type: Literal["tensor"] = "tensor"
    value: list[Any] = []
    dtype: Optional[str] = None

    def to_numpy(self) -> np.ndarray:
        if self.value is None:
            raise ValueError("Tensor is empty")
        if type(self.value) != list:
            raise ValueError("Tensor value is not a list")
        tensor = np.array(self.value, dtype=self.dtype)
        return tensor

    def to_list(self) -> list:
        return self.to_numpy().tolist()

    @staticmethod
    def from_numpy(tensor: np.ndarray, **kwargs):
        return Tensor(value=tensor.tolist(), dtype=str(tensor.dtype), **kwargs)

    @staticmethod
    def from_list(tensor: list, **kwargs):
        return Tensor(value=tensor, **kwargs)


def to_numpy(num: float | int | Tensor) -> np.ndarray:
    if type(num) in (float, int, list):
        return np.array(num)
    elif type(num) == Tensor:
        return num.to_numpy()
    else:
        raise ValueError()


class OutputType(BaseModel):
    """
    This is the base class for all strucutred output types when a node
    wants to return more than one output.
    """

    pass


class ChatConversation(OutputType):
    """
    The result of a chat conversation.
    """

    messages: list[str] = Field(
        default_factory=list, description="The messages in the conversation"
    )
    response: str = Field(default="", description="The response from the chat system")


class ColumnDef(BaseModel):
    name: str
    data_type: (
        Literal["int"] | Literal["float"] | Literal["datetime"] | Literal["object"]
    )


def dtype_name(dtype: str):
    if dtype.startswith("int"):
        return "int"
    if dtype.startswith("float"):
        return "float"
    if dtype.startswith("datetime"):
        return "datetime"
    return "object"


class DataframeRef(AssetRef):
    type: Literal["dataframe"] = "dataframe"
    columns: list[ColumnDef] | None = None
    data: list[list[Any]] | None = None


class TrainTestOutput(OutputType):
    train_X: DataframeRef = DataframeRef()
    train_y: DataframeRef = DataframeRef()
    test_X: DataframeRef = DataframeRef()
    test_y: DataframeRef = DataframeRef()


class RankingResult(BaseType):
    type: Literal["ranking_result"] = "ranking_result"
    score: float
    text: str


class Dataset(OutputType):
    """
    This class represents a dataset, which includes a dataframe of features and a dataframe of targets.
    """

    data: DataframeRef = DataframeRef()
    target: DataframeRef = DataframeRef()


def is_output_type(type):
    try:
        return issubclass(type, OutputType)
    except:
        return False


ModelFileEnums = (
    "comfy.checkpoint_file",
    "comfy.vae_file",
    "comfy.clip_file",
    "comfy.unclip_file",
    "comfy.gligen_file",
    "comfy.clip_vision_file",
    "comfy.control_net_file",
    "comfy.upscale_model_file",
    "comfy.lora_file",
    "comfy.ip_adapter_file",
)


class TypeMetadata(BaseModel):
    """
    Metadata for a type.
    """

    type: str
    optional: bool = False
    values: Optional[list[str | int]] = None
    type_args: list["TypeMetadata"] = []
    type_name: Optional[str] = None

    def __repr__(self):
        return f"TypeMetadata(type={self.type}, optional={self.optional}, values={self.values}, type_args={self.type_args})"

    def is_asset_type(self):
        return self.type in asset_types

    def is_comfy_type(self):
        return self.type.startswith("comfy.")

    def is_enum_type(self):
        return self.type == "enum"

    def is_union_type(self):
        return self.type == "union"

    def get_python_type(self):
        if self.is_enum_type():
            return NameToType[self.type_name]
        else:
            return NameToType[self.type]

    def get_json_schema(self) -> dict[str, Any]:
        """
        Returns a JSON schema for the type.
        """
        if self.type == "any":
            return {}
        if self.is_comfy_type():
            return {"type": "object", "properties": {"id": {"type": "string"}}}
        if self.is_asset_type():
            return {"type": "object", "properties": {"url": {"type": "string"}}}
        if self.type == "none":
            return {"type": "null"}
        if self.type == "int":
            return {"type": "integer"}
        if self.type == "float":
            return {"type": "number"}
        if self.type == "bool":
            return {"type": "boolean"}
        if self.type == "str":
            return {"type": "string"}
        if self.type == "text":
            return {"type": "string"}
        if self.type == "tensor":
            return {"type": "array", "items": {"type": "number"}}
        if self.type == "list":
            return {
                "type": "array",
                "items": self.type_args[0].get_json_schema(),
            }
        if self.type == "dict":
            return {
                "type": "object",
                "properties": {
                    f"key_{i}": t.get_json_schema()
                    for i, t in enumerate(self.type_args)
                },
            }
        if self.type == "union":
            return {
                "anyOf": [t.get_json_schema() for t in self.type_args],
            }
        if self.type == "enum":
            return {
                "type": "string",
                "enum": self.values,
            }
        if self.type in ModelFileEnums:
            return {"type": "object", "properties": {"name": {"type": "string"}}}
        raise ValueError(f"Unknown type: {self.type}")


class OutputSlot(BaseModel):
    """
    An output slot is a slot that can be connected to an input slot.
    """

    type: TypeMetadata
    name: str
    stream: bool = False


def asset_to_ref(asset: Asset):
    if asset.content_type.startswith("image"):
        return ImageRef(asset_id=asset.id)
    elif asset.content_type.startswith("audio"):
        return AudioRef(asset_id=asset.id)
    elif asset.content_type.startswith("video"):
        return VideoRef(asset_id=asset.id)
    elif asset.content_type == "text/csv":
        return DataframeRef(asset_id=asset.id)
    elif asset.content_type.startswith("text"):
        return TextRef(asset_id=asset.id)
    elif asset.content_type.startswith("application/model"):
        return ModelRef(asset_id=asset.id)
    elif asset.content_type.startswith("folder"):
        return AssetRef(asset_id=asset.id)
    else:
        raise ValueError(f"Unknown asset type: {asset.content_type}")


class Message(BaseType):
    type: str = "message"
    id: str | None = None
    thread_id: str | None = None
    user_id: str | None = None
    tool_call_id: str | None = None
    role: str = ""
    name: str = ""
    content: str | list[MessageContent] | None = None
    tool_calls: list[dict] | None = None
    created_at: str | None = None

    @staticmethod
    def from_model(message: MessageModel):
        return Message(
            id=message.id,
            thread_id=message.thread_id,
            tool_call_id=message.tool_call_id,
            role=message.role,
            name=message.name,
            content=message.content,
            tool_calls=message.tool_calls,
            created_at=message.created_at.isoformat(),
        )



from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum


class VideoLlava(ReplicateNode):
    """Video-LLaVA: Learning United Visual Representation by Alignment Before Projection"""

    @classmethod
    def replicate_model_id(cls):
        return "nateraw/video-llava:26387f81b9417278a8578188a31cd763eb3a55ca0f3ec375bf69c713de3fb4e8"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 (Large) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/132630f1-18c6-429e-bc5b-214775997065/video_llava.png",
            "created_at": "2023-11-20T21:16:55.794627Z",
            "description": "Video-LLaVA: Learning United Visual Representation by Alignment Before Projection",
            "github_url": "https://github.com/PKU-YuanGroup/Video-LLaVA",
            "license_url": "https://github.com/PKU-YuanGroup/Video-LLaVA#-license",
            "name": "video-llava",
            "owner": "nateraw",
            "paper_url": "https://arxiv.org/abs/2311.10122",
            "run_count": 280571,
            "url": "https://replicate.com/nateraw/video-llava",
            "visibility": "public",
            "hardware": "Nvidia A40 (Large) GPU",
        }

    @classmethod
    def return_type(cls):
        return str

    image_path: ImageRef = Field(default=ImageRef(), description="Path to image file.")
    video_path: VideoRef = Field(default=VideoRef(), description="Path to video file.")
    text_prompt: str | None = Field(title="Text Prompt", default=None)



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management

class Canny_Edge_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            low_threshold=("INT", {"default": 100, "min": 0, "max": 255, "step": 1}),
            high_threshold=("INT", {"default": 200, "min": 0, "max": 255, "step": 1})
        )

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Line Extractors"

    def execute(self, image, low_threshold, high_threshold, resolution=512, **kwargs):
        from controlnet_aux.canny import CannyDetector

        return (common_annotator_call(CannyDetector(), image, low_threshold=low_threshold, high_threshold=high_threshold, resolution=resolution), )



NODE_CLASS_MAPPINGS = {
    "CannyEdgePreprocessor": Canny_Edge_Preprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "CannyEdgePreprocessor": "Canny Edge"
}


import asyncio
from datetime import datetime
from enum import Enum
import json
from queue import Queue
import urllib.parse
import uuid
import httpx
import joblib
import base64
import PIL.Image
import numpy as np
import pandas as pd

from nodetool.api.types.asset import Asset, AssetCreateRequest, AssetList
from nodetool.api.types.chat import (
    MessageList,
    MessageCreateRequest,
    TaskCreateRequest,
    TaskList,
)
from nodetool.api.types.prediction import PredictionCreateRequest
from nodetool.api.types.workflow import Workflow
from nodetool.common.async_iterators import AsyncByteStream
from nodetool.common.nodetool_api_client import NodetoolAPIClient, Response
from nodetool.metadata.types import Message, Task
from nodetool.workflows.graph import Graph
from nodetool.workflows.types import (
    NodeProgress,
    NodeUpdate,
    ProcessingMessage,
    WorkflowUpdate,
)
from nodetool.workflows.run_job_request import RunJobRequest
from nodetool.metadata.types import (
    AssetRef,
    AudioRef,
    ColumnDef,
    DataframeRef,
    FunctionModel,
    ImageRef,
    ModelRef,
    TextRef,
    VideoRef,
    dtype_name,
)
from nodetool.common.environment import Environment
from nodetool.workflows.base_node import BaseNode
from nodetool.workflows.property import Property
from nodetool.metadata.types import ImageRef
from nodetool.common.environment import Environment


from io import BytesIO
from typing import IO, Any, Literal
from pickle import dumps, loads


log = Environment.get_logger()


class ProcessingContext:
    """
    The processing context is the workflow's interface to the outside world.
    It maintains the state of the workflow and provides methods for interacting with the environment.

    Initialization and State Management:
    - Initializes the context with user ID, authentication token, workflow ID, graph edges, nodes, capabilities, and a message queue.
    - Manages the results of processed nodes and keeps track of processed nodes.
    - Provides methods for popping and posting messages to the message queue.

    Asset Management:
    - Provides methods for finding, downloading, and creating assets (images, audio, text, video, dataframes, models).
    - Handles conversions between different asset formats (e.g., PIL Image to ImageRef, numpy array to ImageRef).
    - Generates presigned URLs for accessing assets.

    API and Storage Integration:
    - Interacts with the Nodetool API client for asset-related operations.
    - Retrieves and manages asset storage and temporary storage instances.
    - Handles file uploads and downloads to/from storage services.

    Workflow Execution:
    - Runs the workflow by sending a RunJobRequest to a remote worker.
    - Processes and handles various types of messages received from the worker (node progress, updates, errors).

    Utility Methods:
    - Provides helper methods for converting values for prediction, handling enums, and parsing S3 URLs.
    - Supports data conversion between different formats (e.g., TextRef to string, DataFrame to pandas DataFrame).

    Attributes:
        user_id (str): The ID of the user running the workflow.
        auth_token (str): The authentication token for the user.
        workflow_id (str): The ID of the workflow being executed.
        capabilities (list[str]): The capabilities required by the workflow.
        graph (Graph): The graph representing the workflow.
        cost (float): The cost of running the workflow.
        results (dict[str, Any]): The results of the processed nodes.
        processed_nodes (set[str]): The IDs of the nodes that have been processed.
        message_queue (Queue | asyncio.Queue): The queue for processing messages.
        api_client (NodetoolAPIClient): The API client for interacting with the Nodetool API.
    """

    user_id: str
    auth_token: str
    workflow_id: str
    capabilities: list[str]
    graph: Graph
    cost: float = 0.0
    results: dict[str, Any]
    processed_nodes: set[str]
    message_queue: Queue | asyncio.Queue
    api_client: NodetoolAPIClient

    def __init__(
        self,
        user_id: str,
        auth_token: str,
        workflow_id: str = "",
        graph: Graph = Graph(),
        queue: Queue | asyncio.Queue | None = None,
        capabilities: list[str] | None = None,
        http_client: httpx.AsyncClient | None = None,
    ):
        self.user_id = user_id
        self.auth_token = auth_token
        self.workflow_id = workflow_id
        self.graph = graph
        self.results = {}
        self.processed_nodes = set()
        self.message_queue = queue if queue else asyncio.Queue()
        self.capabilities = (
            capabilities if capabilities else Environment.get_capabilities()
        )
        self.api_client = Environment.get_nodetool_api_client(self.auth_token)
        self.http_client = (
            httpx.AsyncClient(follow_redirects=True, timeout=600)
            if http_client is None
            else http_client
        )
        assert self.auth_token is not None, "Auth token is required"

    async def pop_message_async(self) -> ProcessingMessage:
        """
        Retrieves and removes a message from the message queue.
        The message queue is used to communicate updates to upstream
        processing.

        Returns:
            The retrieved message from the message queue.
        """
        assert isinstance(self.message_queue, asyncio.Queue)
        return await self.message_queue.get()

    def pop_message(self) -> ProcessingMessage:
        """
        Removes and returns the next message from the message queue.

        Returns:
            The next message from the message queue.
        """
        assert isinstance(self.message_queue, Queue)
        return self.message_queue.get()

    def post_message(self, message: ProcessingMessage):
        """
        Posts a message to the message queue.

        Args:
            message (ProcessingMessage): The message to be posted.
        """
        self.message_queue.put_nowait(message)

    def has_messages(self) -> bool:
        """
        Checks if the processing context has any messages in the message queue.

        Returns:
            bool: True if the message queue is not empty, False otherwise.
        """
        return not self.message_queue.empty()

    def asset_storage_url(self, key: str) -> str:
        """
        Returns the URL of an asset in the asset storage.

        Args:
            key (str): The key of the asset.
        """
        return Environment.get_asset_storage().get_url(key)

    def temp_storage_url(self, key: str) -> str:
        """
        Returns the URL of an asset in the temp storage.

        Args:
            key (str): The key of the asset.
        """
        return Environment.get_temp_storage().get_url(key)

    async def find_asset(self, asset_id: str):
        """
        Finds an asset by id.

        Args:
            asset_id (str): The ID of the asset.

        Returns:
            Asset: The asset with the given ID.
        """
        res = await self.api_client.get(f"api/assets/{asset_id}")
        return Asset(**res.json())

    async def get_asset_url(self, asset_id: str):
        """
        Returns the asset url.

        Args:
            asset_id (str): The ID of the asset.

        Returns:
            str: The URL of the asset.
        """
        asset = await self.find_asset(asset_id)  # type: ignore

        return self.asset_storage_url(asset.file_name)

    def get_result(self, node_id: str, slot: str) -> Any:
        """
        Get the result of a node.

        Results are stored in the context's results dictionary after a node is processed.

        Args:
            node_id (str): The ID of the node.
            slot (str): The slot name.

        Returns:
            Any: The result of the node.
        """
        res = self.results.get(node_id, {})

        if res:
            return res.get(slot, None)
        else:
            return None

    def set_result(self, node_id: str, res: dict[str, Any]):
        """
        Set the result of a node.

        Results are stored in the context's results dictionary after a node is processed.

        Args:
            node_id (str): The ID of the node.
            res (dict[str, Any]): The result of the node.
        """
        self.results[node_id] = res

    def get_node_inputs(self, node_id: str) -> dict[str, Any]:
        """
        Retrieves the inputs for a given node.

        Args:
            node_id (str): The ID of the node.

        Returns:
            dict[str, Any]: A dictionary containing the inputs for the node, where the keys are the input slot names
            and the values are the results from the corresponding source nodes.
        """
        return {
            edge.targetHandle: self.get_result(edge.source, edge.sourceHandle)
            for edge in self.graph.edges
            if edge.target == node_id
        }

    def find_node(self, node_id: str) -> BaseNode:
        """
        Finds a node by its ID.

        Args:
            node_id (str): The ID of the node to be found.

        Returns:
            BaseNode: The node with the given ID.

        Raises:
            ValueError: If the node with the given ID does not exist.
        """
        node = self.graph.find_node(node_id)
        if node is None:
            raise ValueError(f"Node with ID {node_id} does not exist")
        return node

    async def get_workflow(self, workflow_id: str):
        """
        Gets the workflow by ID.
        """
        res = await self.api_client.get(f"api/workflows/{workflow_id}")
        return Workflow(**res.json())

    async def run_prediction(
        self,
        node_id: str,
        provider: Literal["huggingface", "replicate", "openai", "anthropic"],
        model: str,
        params: dict[str, Any] | None = None,
        data: bytes | None = None,
    ) -> Response:
        """
        Run a prediction on a third-party provider.

        Args:
            node_id (str): The ID of the node.
            provider (Literal["huggingface", "replicate", "openai", "anthropic"]): The provider to use for the prediction.
            model (str): The name of the model to use for the prediction.
            params (dict[str, Any] | None, optional): Additional parameters for the prediction. Defaults to None.
            data (bytes | None, optional): Input data for the prediction. Defaults to None.

        Returns:
            Response: The response from the prediction API.

        """

        data_encoded = (
            base64.b64encode(data).decode("utf-8") if isinstance(data, bytes) else None
        )

        res = await self.api_client.post(
            "api/predictions/",
            json=PredictionCreateRequest(
                provider=provider,
                model=model,
                node_id=node_id,
                workflow_id=self.workflow_id if self.workflow_id else "",
                params=params or {},
                data=data_encoded,
            ).model_dump(),
        )
        return res

    async def paginate_assets(
        self,
        parent_id: str | None = None,
        page_size: int = 100,
        cursor: str | None = None,
    ) -> AssetList:
        """
        Lists children assets for a given parent asset.
        Lists top level assets if parent_id is None.

        Args:
            parent_id (str | None, optional): The ID of the parent asset. Defaults to None.
            page_size (int, optional): The number of assets to return. Defaults to 100.
            cursor (str | None, optional): The cursor for pagination. Defaults to None.

        Returns:
            AssetList: The list of assets.
        """
        res = await self.api_client.get(
            "api/assets/",
            params={"parent_id": parent_id, page_size: page_size, "cursor": cursor},
        )
        return AssetList(**res.json())

    async def refresh_uri(self, asset: AssetRef):
        """
        Refreshes the URI of the asset.

        Args:
            asset (AssetRef): The asset to refresh.
        """
        if asset.asset_id:
            asset.uri = await self.get_asset_url(asset.asset_id)
        elif asset.temp_id:
            asset.uri = self.temp_storage_url(asset.temp_id)

    async def create_asset(
        self,
        name: str,
        content_type: str,
        content: IO,
        parent_id: str | None = None,
    ) -> Asset:
        """
        Creates an asset with the given name, content type, content, and optional parent ID.

        Args:
            name (str): The name of the asset.
            content_type (str): The content type of the asset.
            content (IO): The content of the asset.
            parent_id (str | None, optional): The ID of the parent asset. Defaults to None.

        Returns:
            Asset: The created asset.

        """
        content.seek(0)

        req = AssetCreateRequest(
            workflow_id=self.workflow_id,
            name=name,
            content_type=content_type,
            parent_id=parent_id,
        )
        res = await self.api_client.post(
            "api/assets/",
            data={"json": req.model_dump_json()},
            files={"file": (name, content, content_type)},
        )

        return Asset(**res.json())

    async def create_temp_asset(self, content: IO, ext: str = "") -> AssetRef:
        """
        Uploads a temporary asset with the given content and extension.

        Args:
            content (IO): The content of the asset.
            ext (str, optional): The extension of the asset. Defaults to "".

        Returns:
            str: The URL of the created temporary asset.
        """
        key = uuid.uuid4().hex
        if ext != "":
            key += "." + ext
        url = self.temp_storage_url(key)
        # parse uri
        path = urllib.parse.urlparse(url).path
        stream = AsyncByteStream(content.read())
        await self.api_client.put(path, content=stream)
        return AssetRef(uri=url, temp_id=key)

    async def create_message(self, req: MessageCreateRequest):
        """
        Creates a message for a thread.

        Args:
            req (MessageCreateRequest): The message to create.

        Returns:
            Message: The created message.
        """
        res = await self.api_client.post("api/messages/", json=req.model_dump())
        return Message(**res.json())

    async def get_messages(
        self,
        thread_id: str,
        limit: int = 10,
        start_key: str | None = None,
        reverse: bool = False,
    ):
        """
        Gets messages for a thread.

        Args:
            thread_id (str): The ID of the thread.
            limit (int, optional): The number of messages to return. Defaults to 10.
            start_key (str, optional): The start key for pagination. Defaults to None.
            reverse (bool, optional): Whether to reverse the order of messages. Defaults to False.

        Returns:
            MessageList: The list of messages.
        """
        res = await self.api_client.get(
            "api/messages/",
            params={"thread_id": thread_id, "limit": limit, "cursor": start_key},
        )
        return MessageList(**res.json())

    async def create_task(self, task: TaskCreateRequest):
        """
        Creates a task.

        Args:
            task (TaskCreateRequest): The task to create.

        Returns:
            Task: The created task.
        """
        res = await self.api_client.post("api/tasks/", json=task.model_dump())
        return Task(**res.json())

    async def get_tasks(self, thread_id: str):
        """
        Gets tasks for a thread.

        Args:
            thread_id (str): The ID of the thread.

        Returns:
            TaskList: The list of tasks.
        """
        res = await self.api_client.get("api/tasks/", params={"thread_id": thread_id})
        return TaskList(**res.json())

    async def download_asset(self, asset_id: str) -> IO:
        """
        Downloads an asset from the asset storage api.

        Args:
            asset_id (str): The ID of the asset to download.

        Returns:
            IO: The downloaded asset.
        """
        asset = await self.find_asset(asset_id)
        io = BytesIO()
        await Environment.get_asset_storage().download(asset.file_name, io)
        io.seek(0)
        return io

    async def download_temp_asset(self, temp_id: str) -> IO:
        """
        Downloads a temporary asset from the temp storage.

        Args:
            temp_id (str): The ID of the temporary asset to download.

        Returns:
            IO: The downloaded temporary asset.
        """
        io = BytesIO()
        await Environment.get_temp_storage().download(temp_id, io)
        io.seek(0)
        return io

    async def http_get(self, url: str) -> bytes:
        """
        Sends an HTTP GET request to the specified URL.

        Args:
            url (str): The URL to send the request to.

        Returns:
            bytes: The response content.
        """
        response = await self.http_client.get(url)
        response.raise_for_status()
        return response.content

    async def download_file(self, url: str) -> IO:
        """
        Download a file from URL.

        Args:
            url (str): The URL of the file to download.

        Returns:
            IO: The downloaded file.
        """
        from nodetool.common.encoding import decode_bytes_io

        url_parsed = urllib.parse.urlparse(url)

        if url_parsed.scheme == "data":
            file = decode_bytes_io(url.split(",")[1])
            # parse file ext from data uri
            ext = url.split(",")[0].split(";")[0].split("/")[1]
            file.name = f"{uuid.uuid4()}.{ext}"
            return file

        return BytesIO(await self.http_get(url))

    async def asset_to_io(self, asset_ref: AssetRef) -> IO:
        """
        Converts an AssetRef object to an IO object.

        Args:
            asset_ref (AssetRef): The AssetRef object to convert.

        Returns:
            IO: The converted IO object.

        Raises:
            ValueError: If the AssetRef is empty.
        """
        # Asset ID takes precedence over URI
        # as the URI could be expired
        if asset_ref.asset_id is not None:
            return await self.download_asset(asset_ref.asset_id)
        elif asset_ref.temp_id is not None:
            return await self.download_temp_asset(asset_ref.temp_id)
        elif asset_ref.uri != "":
            return await self.download_file(asset_ref.uri)
        raise ValueError(f"AssetRef is empty {asset_ref}")

    async def image_to_pil(self, image_ref: ImageRef) -> PIL.Image.Image:
        """
        Converts the image to a PIL Image object.

        Args:
            context (ProcessingContext): The processing context.
        """
        buffer = await self.asset_to_io(image_ref)
        return PIL.Image.open(buffer).convert("RGB")

    async def image_to_base64(self, image_ref: ImageRef) -> str:
        """
        Converts the image to a base64-encoded string.

        Args:
            context (ProcessingContext): The processing context.
        """
        buffer = await self.asset_to_io(image_ref)
        return base64.b64encode(buffer.read()).decode("utf-8")

    async def audio_to_audio_segment(self, audio_ref: AudioRef):
        """
        Converts the audio to an AudioSegment object.

        Args:
            context (ProcessingContext): The processing context.
        """
        import pydub

        audio_bytes = await self.asset_to_io(audio_ref)
        return pydub.AudioSegment.from_file(audio_bytes)

    async def audio_from_io(
        self,
        buffer: IO,
        name: str | None = None,
        parent_id: str | None = None,
    ) -> "AudioRef":
        """
        Creates an AudioRef from an IO object.

        Args:
            buffer (IO): The IO object.
            name (Optional[str], optional): The name of the asset. Defaults to None
            parent_id (Optional[str], optional): The parent ID of the asset. Defaults to None.

        Returns:
            AudioRef: The AudioRef object.
        """
        if name:
            asset = await self.create_asset(
                name=name, content_type="audio/mp3", content=buffer, parent_id=parent_id
            )
            return AudioRef(asset_id=asset.id, uri=asset.get_url or "")
        else:
            ref = await self.create_temp_asset(buffer, "mp3")
            return AudioRef(temp_id=ref.temp_id, uri=ref.uri)

    async def audio_from_bytes(
        self,
        b: bytes,
        name: str | None = None,
        parent_id: str | None = None,
    ) -> "AudioRef":
        """
        Creates an AudioRef from a bytes object.

        Args:
            context (ProcessingContext): The processing context.
            b (bytes): The bytes object.
            name (Optional[str], optional): The name of the asset. Defaults to None.

        Returns:
            AudioRef: The AudioRef object.
        """
        return await self.audio_from_io(BytesIO(b), name=name, parent_id=parent_id)

    async def audio_to_numpy(self, audio_ref: AudioRef) -> tuple[np.ndarray, int]:
        audio_segment = await self.audio_to_audio_segment(audio_ref)
        return np.array(audio_segment.get_array_of_samples()), audio_segment.frame_rate

    async def audio_from_segment(
        self,
        audio_segment: "pydub.AudioSegment",  # type: ignore
        name: str | None = None,
        parent_id: str | None = None,
        **kwargs,
    ) -> "AudioRef":
        """
        Converts an audio segment to an AudioRef object.

        Args:
            audio_segment (pydub.AudioSegment): The audio segment to convert.
            name (str, optional): The name of the audio. Defaults to None.
            parent_id (str, optional): The ID of the parent asset. Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            AudioRef: The converted AudioRef object.

        """
        buffer = BytesIO()
        audio_segment.export(buffer, format="mp3")
        buffer.seek(0)
        if name:
            asset = await self.create_asset(
                name, "audio/mp3", buffer, parent_id=parent_id
            )
            return AudioRef(asset_id=asset.id, uri=asset.get_url or "")
        else:
            ref = await self.create_temp_asset(buffer, "mp3")
            return AudioRef(
                temp_id=ref.temp_id,
                uri=ref.uri,
            )

    async def dataframe_to_pandas(self, df: DataframeRef) -> "pd.DataFrame":
        """
        Converts a DataframeRef object to a pandas DataFrame.

        Args:
            df (DataframeRef): The DataframeRef object to convert.

        Returns:
            pandas.DataFrame: The converted pandas DataFrame.
        """
        import pandas as pd

        if df.columns:
            column_names = [col.name for col in df.columns]
            return pd.DataFrame(df.data, columns=column_names)
        else:
            io = await self.asset_to_io(df)
            df = loads(io.read())
            assert isinstance(df, pd.DataFrame), "Is not a dataframe"
            return df

    async def dataframe_from_pandas(
        self, data: "pd.DataFrame", name: str | None = None
    ) -> "DataframeRef":
        """
        Converts a pandas DataFrame to a DataframeRef object.

        Args:
            data (pd.DataFrame): The pandas DataFrame to convert.
            name (str | None, optional): The name of the asset. Defaults to None.

        Returns:
            DataframeRef: The converted DataframeRef object.
        """
        buffer = BytesIO(dumps(data))
        if name:
            asset = await self.create_asset(name, "application/octet-stream", buffer)
            return DataframeRef(asset_id=asset.id, uri=asset.get_url or "")
        else:
            ref = await self.create_temp_asset(buffer)
            # TODO: avoid for large tables
            rows = data.values.tolist()
            column_defs = [
                ColumnDef(name=name, data_type=dtype_name(dtype.name))
                for name, dtype in zip(data.columns, data.dtypes)
            ]
            return DataframeRef(temp_id=ref.temp_id, columns=column_defs, data=rows)

    async def image_from_io(
        self,
        buffer: IO,
        name: str | None = None,
        parent_id: str | None = None,
    ) -> "ImageRef":
        """
        Creates an ImageRef from an IO object.

        Args:
            buffer (IO): The IO object.
            name (Optional[str], optional): The name of the asset. Defaults to None
            parent_id (Optional[str], optional): The parent ID of the asset. Defaults to None.

        Returns:
            ImageRef: The ImageRef object.
        """
        if name:
            asset = await self.create_asset(
                name=name, content_type="image/png", content=buffer, parent_id=parent_id
            )
            return ImageRef(asset_id=asset.id, uri=asset.get_url or "")
        else:
            ref = await self.create_temp_asset(buffer)
            return ImageRef(temp_id=ref.temp_id, uri=ref.uri)

    async def image_from_url(
        self,
        url: str,
        name: str | None = None,
        parent_id: str | None = None,
    ) -> "ImageRef":
        """
        Creates an ImageRef from a URL.

        Args:
            context (ProcessingContext): The processing context.
            url (str): The URL.
            name (Optional[str], optional): The name of the asset. Defaults to None.

        Returns:
            ImageRef: The ImageRef object.
        """
        return await self.image_from_io(
            await self.download_file(url), name=name, parent_id=parent_id
        )

    async def image_from_bytes(
        self,
        b: bytes,
        name: str | None = None,
        parent_id: str | None = None,
    ) -> "ImageRef":
        """
        Creates an ImageRef from a bytes object.

        Args:
            context (ProcessingContext): The processing context.
            b (bytes): The bytes object.
            name (Optional[str], optional): The name of the asset. Defaults to None.

        Returns:
            ImageRef: The ImageRef object.
        """
        return await self.image_from_io(BytesIO(b), name=name, parent_id=parent_id)

    async def image_from_base64(
        self,
        b64: str,
        name: str | None = None,
        parent_id: str | None = None,
    ) -> "ImageRef":
        """
        Creates an ImageRef from a base64-encoded string.

        Args:
            context (ProcessingContext): The processing context.
            b64 (str): The base64-encoded string.
            name (Optional[str], optional): The name of the asset. Defaults to None.

        Returns:
            ImageRef: The ImageRef object.
        """
        return await self.image_from_bytes(
            base64.b64decode(b64), name=name, parent_id=parent_id
        )

    async def image_from_pil(
        self,
        image: PIL.Image.Image,
        name: str | None = None,
        parent_id: str | None = None,
    ) -> "ImageRef":
        """
        Creates an ImageRef from a PIL Image object.

        Args:
            image (Image.Image): The PIL Image object.
            name (Optional[str], optional): The name of the asset. Defaults to None.
            parent_id (Optional[str], optional): The parent ID of the asset. Defaults to None.

        Returns:
            ImageRef: The ImageRef object.
        """
        buffer = BytesIO()
        image.save(buffer, format="png")
        buffer.seek(0)
        return await self.image_from_io(buffer, name=name, parent_id=parent_id)

    async def image_from_numpy(
        self, image: np.ndarray, name: str | None = None, parent_id: str | None = None
    ) -> "ImageRef":
        """
        Creates an ImageRef from a numpy array.

        Args:
            image (np.ndarray): The numpy array.
            name (Optional[str], optional): The name of the asset. Defaults to None.
            parent_id (Optional[str], optional): The parent ID of the asset. Defaults to None.

        Returns:
            ImageRef: The ImageRef object.
        """
        return await self.image_from_pil(PIL.Image.fromarray(image), name=name)

    async def to_str(self, text_ref: TextRef | str) -> str:
        """
        Converts a TextRef to a string.

        Args:
            text_ref (TextRef): The TextRef object.

        Returns:
            str: The string.
        """
        if isinstance(text_ref, TextRef):
            stream = await self.asset_to_io(text_ref)
            return stream.read().decode("utf-8")
        else:
            return text_ref

    async def text_from_str(
        self,
        s: str,
        name: str | None = None,
        content_type: str = "text/plain",
        parent_id: str | None = None,
    ) -> "TextRef":
        buffer = BytesIO(s.encode("utf-8"))
        if name:
            asset = await self.create_asset(
                name, content_type, buffer, parent_id=parent_id
            )
            return TextRef(asset_id=asset.id, uri=asset.get_url or "")
        else:
            ref = await self.create_temp_asset(buffer)
            return TextRef(temp_id=ref.temp_id, uri=ref.uri)

    async def video_from_io(
        self,
        buffer: IO,
        name: str | None = None,
        parent_id: str | None = None,
    ):
        """
        Creates an VideoRef from an IO object.

        Args:
            context (ProcessingContext): The processing context.
            buffer (IO): The IO object.
            name (Optional[str], optional): The name of the asset. Defaults to None.

        Returns:
            VideoRef: The VideoRef object.
        """
        if name:
            asset = await self.create_asset(
                name, "video/mpeg", buffer, parent_id=parent_id
            )
            return VideoRef(asset_id=asset.id, uri=asset.get_url or "")
        else:
            ref = await self.create_temp_asset(buffer)
            return VideoRef(temp_id=ref.temp_id, uri=ref.uri)

    async def to_estimator(self, model_ref: ModelRef):
        """
        Converts a model reference to an estimator object.

        Args:
            model_ref (ModelRef): The model reference to convert.

        Returns:
            The loaded estimator object.

        Raises:
            ValueError: If the model reference is empty.
        """
        if model_ref.asset_id is None:
            raise ValueError("ModelRef is empty")
        file = await self.asset_to_io(model_ref)
        return joblib.load(file)

    async def from_estimator(self, est: "BaseEstimator", **kwargs):  # type: ignore
        """
        Create a model asset from an estimator.

        Args:
            est (BaseEstimator): The estimator object to be serialized.
            **kwargs: Additional keyword arguments.

        Returns:
            ModelRef: A reference to the created model asset.

        """
        stream = BytesIO()
        joblib.dump(est, stream)
        stream.seek(0)
        asset = await self.create_asset("model", "application/model", stream)

        return ModelRef(uri=asset.get_url or "", asset_id=asset.id, **kwargs)

    async def convert_value_for_prediction(
        self,
        property: Property,
        value: Any,
    ):
        """
        Converts the property value for a remote api prediction on replicate or huggingface.

        Args:
            property (Property): The property.
            value: The value to be converted.

        Raises:
            NotImplementedError: If the self type is 'tensor' and not implemented.
            ValueError: If the self value is an invalid enum value.
        """

        if isinstance(value, AssetRef):
            if value.is_empty():
                return None
            io = await self.asset_to_io(value)
            if isinstance(value, TextRef):
                return io.read().decode("utf-8")
            else:
                img_bytes = io.read()
                b64 = base64.b64encode(img_bytes).decode("utf-8")
                if isinstance(value, ImageRef):
                    return "data:image/png;base64," + b64
                elif isinstance(value, AudioRef):
                    return "data:audio/mp3;base64," + b64
                elif isinstance(value, VideoRef):
                    return "data:video/mp4;base64," + b64
                else:
                    return b64
        elif property.type.type == "tensor":
            raise NotImplementedError()
        elif property.type.type == "enum":
            if value is None:
                return None
            elif isinstance(value, str):
                return value
            elif isinstance(value, Enum):
                return value.value
            else:
                raise ValueError(f"Invalid enum value {value} : {type(value)}")
        else:
            return value

    async def image_from_openai(self, file_id: str) -> "ImageRef":
        """
        Creates an ImageRef from an OpenAI ImageFile object.

        Args:
            file_id (str): The ID of the file.

        Returns:
            ImageRef: The ImageRef object.
        """
        client = Environment.get_openai_client()
        res = await client.files.content(file_id)
        return await self.image_from_bytes(res.content)

    async def run_worker(self, req: RunJobRequest) -> dict[str, Any]:
        """
        Runs the workflow using the provided graph and parameters.

        Args:
            graph (Graph): The graph representing the workflow.
            params (dict, optional): Additional parameters for the workflow. Defaults to {}.

        Returns:
            dict[str, Any]: The result of running the workflow.
        """

        # TODO: logic to determine which worker to use for given workflow

        url = Environment.get_worker_url()
        assert url is not None, "Worker URL is required"
        assert self.auth_token is not None, "Auth token is required"

        req.auth_token = self.auth_token
        req.user_id = self.user_id

        headers = {
            "Content-Type": "application/json",
        }

        if req.env is None:
            req.env = {}

        if Environment.get_use_ngrok():
            api_tunnel_url = Environment.get_api_tunnel_url()
            req.env["NODETOOL_API_URL"] = api_tunnel_url
            req.env["S3_ENDPOINT_URL"] = f"{api_tunnel_url}/storage"

        log.info("===== Run remote worker ====")
        log.info(url)
        # log.info(json.dumps(req.model_dump(), indent=2))
        result = {}

        async with self.http_client.stream(
            "POST",
            url,
            headers=headers,
            json=req.model_dump(),
        ) as response:
            response.raise_for_status()
            async for line in response.aiter_lines():
                try:
                    message = json.loads(line)
                    print(message)

                    if isinstance(message, dict) and "type" in message:
                        if message["type"] == "node_progress":
                            self.post_message(NodeProgress(**message))

                        elif message["type"] == "node_update":
                            self.post_message(NodeUpdate(**message))

                        elif message["type"] == "workflow_update":
                            self.post_message(WorkflowUpdate(**message))

                        elif message["type"] == "error":
                            raise Exception(message["error"])

                except json.JSONDecodeError as e:
                    log.error("Error decoding message: " + str(e))

        return result

    def get_chroma_client(self):
        import chromadb
        from chromadb.config import DEFAULT_DATABASE, DEFAULT_TENANT

        settings = Environment.get_chroma_settings()

        if Environment.get_chroma_url():
            admin = chromadb.AdminClient()
            tenant = f"tenant_{self.user_id}"
            try:
                admin.get_tenant(tenant)
            except Exception:
                admin.create_tenant(tenant)
                admin.create_database(DEFAULT_DATABASE, tenant)
            return chromadb.HttpClient(
                host=Environment.get_chroma_url(), settings=settings
            )
        else:
            return chromadb.PersistentClient(
                path=Environment.get_chroma_path(),
                tenant=DEFAULT_TENANT,
                database=DEFAULT_DATABASE,
            )

    def load_llama_model(self, name, **kwargs):
        from llama_cpp import Llama
        from llama_cpp.llama_tokenizer import LlamaHFTokenizer

        model = Environment.find_llama_model(name)
        assert model, "model not found"

        if model.local_path is None:
            raise Exception("model path is None")

        if not model.local_path.exists():
            raise Exception(f"path doesn not exist {model.local_path}")

        return Llama(
            verbose=True,
            model_path=str(model.local_path),
            chat_format="functionary-v2" if isinstance(model, FunctionModel) else None,
            # tokenizer=LlamaHFTokenizer.from_pretrained(found_model.repo_id),
            **kwargs,
        )



# Copyright (c) Facebook, Inc. and its affiliates.
import logging
import numpy as np
from typing import Dict, List, Optional, Tuple
import torch
from torch import nn

from custom_detectron2.config import configurable
from custom_detectron2.data.detection_utils import convert_image_to_rgb
from custom_detectron2.layers import move_device_like
from custom_detectron2.structures import ImageList, Instances
from custom_detectron2.utils.events import get_event_storage
from custom_detectron2.utils.logger import log_first_n

from ..backbone import Backbone, build_backbone
from ..postprocessing import detector_postprocess
from ..proposal_generator import build_proposal_generator
from ..roi_heads import build_roi_heads
from .build import META_ARCH_REGISTRY

__all__ = ["GeneralizedRCNN", "ProposalNetwork"]


@META_ARCH_REGISTRY.register()
class GeneralizedRCNN(nn.Module):
    """
    Generalized R-CNN. Any models that contains the following three components:
    1. Per-image feature extraction (aka backbone)
    2. Region proposal generation
    3. Per-region feature extraction and prediction
    """

    @configurable
    def __init__(
        self,
        *,
        backbone: Backbone,
        proposal_generator: nn.Module,
        roi_heads: nn.Module,
        pixel_mean: Tuple[float],
        pixel_std: Tuple[float],
        input_format: Optional[str] = None,
        vis_period: int = 0,
    ):
        """
        Args:
            backbone: a backbone module, must follow detectron2's backbone interface
            proposal_generator: a module that generates proposals using backbone features
            roi_heads: a ROI head that performs per-region computation
            pixel_mean, pixel_std: list or tuple with #channels element, representing
                the per-channel mean and std to be used to normalize the input image
            input_format: describe the meaning of channels of input. Needed by visualization
            vis_period: the period to run visualization. Set to 0 to disable.
        """
        super().__init__()
        self.backbone = backbone
        self.proposal_generator = proposal_generator
        self.roi_heads = roi_heads

        self.input_format = input_format
        self.vis_period = vis_period
        if vis_period > 0:
            assert input_format is not None, "input_format is required for visualization!"

        self.register_buffer("pixel_mean", torch.tensor(pixel_mean).view(-1, 1, 1), False)
        self.register_buffer("pixel_std", torch.tensor(pixel_std).view(-1, 1, 1), False)
        assert (
            self.pixel_mean.shape == self.pixel_std.shape
        ), f"{self.pixel_mean} and {self.pixel_std} have different shapes!"

    @classmethod
    def from_config(cls, cfg):
        backbone = build_backbone(cfg)
        return {
            "backbone": backbone,
            "proposal_generator": build_proposal_generator(cfg, backbone.output_shape()),
            "roi_heads": build_roi_heads(cfg, backbone.output_shape()),
            "input_format": cfg.INPUT.FORMAT,
            "vis_period": cfg.VIS_PERIOD,
            "pixel_mean": cfg.MODEL.PIXEL_MEAN,
            "pixel_std": cfg.MODEL.PIXEL_STD,
        }

    @property
    def device(self):
        return self.pixel_mean.device

    def _move_to_current_device(self, x):
        return move_device_like(x, self.pixel_mean)

    def visualize_training(self, batched_inputs, proposals):
        """
        A function used to visualize images and proposals. It shows ground truth
        bounding boxes on the original image and up to 20 top-scoring predicted
        object proposals on the original image. Users can implement different
        visualization functions for different models.

        Args:
            batched_inputs (list): a list that contains input to the model.
            proposals (list): a list that contains predicted proposals. Both
                batched_inputs and proposals should have the same length.
        """
        from custom_detectron2.utils.visualizer import Visualizer

        storage = get_event_storage()
        max_vis_prop = 20

        for input, prop in zip(batched_inputs, proposals):
            img = input["image"]
            img = convert_image_to_rgb(img.permute(1, 2, 0), self.input_format)
            v_gt = Visualizer(img, None)
            v_gt = v_gt.overlay_instances(boxes=input["instances"].gt_boxes)
            anno_img = v_gt.get_image()
            box_size = min(len(prop.proposal_boxes), max_vis_prop)
            v_pred = Visualizer(img, None)
            v_pred = v_pred.overlay_instances(
                boxes=prop.proposal_boxes[0:box_size].tensor.cpu().numpy()
            )
            prop_img = v_pred.get_image()
            vis_img = np.concatenate((anno_img, prop_img), axis=1)
            vis_img = vis_img.transpose(2, 0, 1)
            vis_name = "Left: GT bounding boxes;  Right: Predicted proposals"
            storage.put_image(vis_name, vis_img)
            break  # only visualize one image in a batch

    def forward(self, batched_inputs: List[Dict[str, torch.Tensor]]):
        """
        Args:
            batched_inputs: a list, batched outputs of :class:`DatasetMapper` .
                Each item in the list contains the inputs for one image.
                For now, each item in the list is a dict that contains:

                * image: Tensor, image in (C, H, W) format.
                * instances (optional): groundtruth :class:`Instances`
                * proposals (optional): :class:`Instances`, precomputed proposals.

                Other information that's included in the original dicts, such as:

                * "height", "width" (int): the output resolution of the model, used in inference.
                  See :meth:`postprocess` for details.

        Returns:
            list[dict]:
                Each dict is the output for one input image.
                The dict contains one key "instances" whose value is a :class:`Instances`.
                The :class:`Instances` object has the following keys:
                "pred_boxes", "pred_classes", "scores", "pred_masks", "pred_keypoints"
        """
        if not self.training:
            return self.inference(batched_inputs)

        images = self.preprocess_image(batched_inputs)
        if "instances" in batched_inputs[0]:
            gt_instances = [x["instances"].to(self.device) for x in batched_inputs]
        else:
            gt_instances = None

        features = self.backbone(images.tensor)

        if self.proposal_generator is not None:
            proposals, proposal_losses = self.proposal_generator(images, features, gt_instances)
        else:
            assert "proposals" in batched_inputs[0]
            proposals = [x["proposals"].to(self.device) for x in batched_inputs]
            proposal_losses = {}

        _, detector_losses = self.roi_heads(images, features, proposals, gt_instances)
        if self.vis_period > 0:
            storage = get_event_storage()
            if storage.iter % self.vis_period == 0:
                self.visualize_training(batched_inputs, proposals)

        losses = {}
        losses.update(detector_losses)
        losses.update(proposal_losses)
        return losses

    def inference(
        self,
        batched_inputs: List[Dict[str, torch.Tensor]],
        detected_instances: Optional[List[Instances]] = None,
        do_postprocess: bool = True,
    ):
        """
        Run inference on the given inputs.

        Args:
            batched_inputs (list[dict]): same as in :meth:`forward`
            detected_instances (None or list[Instances]): if not None, it
                contains an `Instances` object per image. The `Instances`
                object contains "pred_boxes" and "pred_classes" which are
                known boxes in the image.
                The inference will then skip the detection of bounding boxes,
                and only predict other per-ROI outputs.
            do_postprocess (bool): whether to apply post-processing on the outputs.

        Returns:
            When do_postprocess=True, same as in :meth:`forward`.
            Otherwise, a list[Instances] containing raw network outputs.
        """
        assert not self.training

        images = self.preprocess_image(batched_inputs)
        features = self.backbone(images.tensor)

        if detected_instances is None:
            if self.proposal_generator is not None:
                proposals, _ = self.proposal_generator(images, features, None)
            else:
                assert "proposals" in batched_inputs[0]
                proposals = [x["proposals"].to(self.device) for x in batched_inputs]

            results, _ = self.roi_heads(images, features, proposals, None)
        else:
            detected_instances = [x.to(self.device) for x in detected_instances]
            results = self.roi_heads.forward_with_given_boxes(features, detected_instances)

        if do_postprocess:
            assert not torch.jit.is_scripting(), "Scripting is not supported for postprocess."
            return GeneralizedRCNN._postprocess(results, batched_inputs, images.image_sizes)
        return results

    def preprocess_image(self, batched_inputs: List[Dict[str, torch.Tensor]]):
        """
        Normalize, pad and batch the input images.
        """
        images = [self._move_to_current_device(x["image"]) for x in batched_inputs]
        images = [(x - self.pixel_mean) / self.pixel_std for x in images]
        images = ImageList.from_tensors(
            images,
            self.backbone.size_divisibility,
            padding_constraints=self.backbone.padding_constraints,
        )
        return images

    @staticmethod
    def _postprocess(instances, batched_inputs: List[Dict[str, torch.Tensor]], image_sizes):
        """
        Rescale the output instances to the target size.
        """
        # note: private function; subject to changes
        processed_results = []
        for results_per_image, input_per_image, image_size in zip(
            instances, batched_inputs, image_sizes
        ):
            height = input_per_image.get("height", image_size[0])
            width = input_per_image.get("width", image_size[1])
            r = detector_postprocess(results_per_image, height, width)
            processed_results.append({"instances": r})
        return processed_results


@META_ARCH_REGISTRY.register()
class ProposalNetwork(nn.Module):
    """
    A meta architecture that only predicts object proposals.
    """

    @configurable
    def __init__(
        self,
        *,
        backbone: Backbone,
        proposal_generator: nn.Module,
        pixel_mean: Tuple[float],
        pixel_std: Tuple[float],
    ):
        """
        Args:
            backbone: a backbone module, must follow detectron2's backbone interface
            proposal_generator: a module that generates proposals using backbone features
            pixel_mean, pixel_std: list or tuple with #channels element, representing
                the per-channel mean and std to be used to normalize the input image
        """
        super().__init__()
        self.backbone = backbone
        self.proposal_generator = proposal_generator
        self.register_buffer("pixel_mean", torch.tensor(pixel_mean).view(-1, 1, 1), False)
        self.register_buffer("pixel_std", torch.tensor(pixel_std).view(-1, 1, 1), False)

    @classmethod
    def from_config(cls, cfg):
        backbone = build_backbone(cfg)
        return {
            "backbone": backbone,
            "proposal_generator": build_proposal_generator(cfg, backbone.output_shape()),
            "pixel_mean": cfg.MODEL.PIXEL_MEAN,
            "pixel_std": cfg.MODEL.PIXEL_STD,
        }

    @property
    def device(self):
        return self.pixel_mean.device

    def _move_to_current_device(self, x):
        return move_device_like(x, self.pixel_mean)

    def forward(self, batched_inputs):
        """
        Args:
            Same as in :class:`GeneralizedRCNN.forward`

        Returns:
            list[dict]:
                Each dict is the output for one input image.
                The dict contains one key "proposals" whose value is a
                :class:`Instances` with keys "proposal_boxes" and "objectness_logits".
        """
        images = [self._move_to_current_device(x["image"]) for x in batched_inputs]
        images = [(x - self.pixel_mean) / self.pixel_std for x in images]
        images = ImageList.from_tensors(
            images,
            self.backbone.size_divisibility,
            padding_constraints=self.backbone.padding_constraints,
        )
        features = self.backbone(images.tensor)

        if "instances" in batched_inputs[0]:
            gt_instances = [x["instances"].to(self.device) for x in batched_inputs]
        elif "targets" in batched_inputs[0]:
            log_first_n(
                logging.WARN, "'targets' in the model inputs is now renamed to 'instances'!", n=10
            )
            gt_instances = [x["targets"].to(self.device) for x in batched_inputs]
        else:
            gt_instances = None
        proposals, proposal_losses = self.proposal_generator(images, features, gt_instances)
        # In training, the proposals are not useful at all but we generate them anyway.
        # This makes RPN-only models about 5% slower.
        if self.training:
            return proposal_losses

        processed_results = []
        for results_per_image, input_per_image, image_size in zip(
            proposals, batched_inputs, images.image_sizes
        ):
            height = input_per_image.get("height", image_size[0])
            width = input_per_image.get("width", image_size[1])
            r = detector_postprocess(results_per_image, height, width)
            processed_results.append({"proposals": r})
        return processed_results



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.openai.image import Size
from nodetool.nodes.openai.image import Quality
from nodetool.nodes.openai.image import Style

class Dall_E(GraphNode):
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The prompt to use.')
    size: Size | GraphNode | tuple[GraphNode, str] = Field(default=Size('1024x1024'), description='The size of the image to generate.')
    quality: Quality | GraphNode | tuple[GraphNode, str] = Field(default=Quality('standard'), description='The quality of the image to generate.')
    style: Style | GraphNode | tuple[GraphNode, str] = Field(default=Style('natural'), description='The style to use.')
    @classmethod
    def get_node_type(cls): return "openai.image.Dall_E"





import base64
import io
from typing import Optional

import PIL.Image
import pandas as pd
from pydub import AudioSegment


def decode_bytes_io(encoded: str) -> io.BytesIO:
    """
    base64 decode an image.
    """
    image_bytes = base64.b64decode(encoded)
    bytes_io = io.BytesIO(image_bytes)
    return bytes_io


def encode_bytes(b: bytes) -> str:
    """
    base64 encode bytes
    """
    return base64.b64encode(b).decode("utf-8")


def encode_image(image: PIL.Image.Image, format: str = "png") -> str:
    """
    base64 encode an image.
    """
    buffered = io.BytesIO()
    image.save(buffered, format=format)
    return base64.b64encode(buffered.getvalue()).decode("utf-8")


def decode_image(image: str) -> PIL.Image.Image:
    """
    base64 decode an image.
    """
    image_bytes = base64.b64decode(image)
    return PIL.Image.open(io.BytesIO(image_bytes))


def encode_audio(audio: AudioSegment, format: str = "mp3") -> str:
    """
    base64 encode an audio.
    """
    buffered = io.BytesIO()
    audio.export(buffered, format=format)
    return base64.b64encode(buffered.getvalue()).decode("utf-8")


def decode_audio(audio: str) -> AudioSegment:
    """
    base64 decode an audio.
    """
    audio_bytes = base64.b64decode(audio)
    return AudioSegment.from_file(io.BytesIO(audio_bytes), format="mp3")


# def encode_latents(
#     vae: AutoencoderKL, seed: int, init_image: PIL.Image.Image
# ) -> torch.Tensor:
#     """
#     Encode the latents for the given image.
#     """
#     generator = torch.Generator().manual_seed(seed)

#     vae_scale_factor = 2 ** (len(vae.config.block_out_channels) - 1)  # type: ignore
#     image_processor = VaeImageProcessor(vae_scale_factor=vae_scale_factor)

#     image = image_processor.preprocess(init_image)
#     init_latents = vae.encode(torch.FloatTensor(image)).latent_dist.sample(generator)
#     init_latents = vae.config.scaling_factor * init_latents  # type: ignore
#     return init_latents


# def decode_latents(vae: AutoencoderKL, latents: torch.Tensor) -> PIL.Image.Image:
#     """
#     Decode the latents to an image.
#     """
#     latents = latents.to(torch.float32).to("cpu").unsqueeze(0)
#     image_tensor = vae.decode(latents / vae.config.scaling_factor, return_dict=False)[0]  # type: ignore
#     vae_scale_factor = 2 ** (len(vae.config.block_out_channels) - 1)  # type: ignore
#     image_processor = VaeImageProcessor(vae_scale_factor=vae_scale_factor)

#     image = image_processor.postprocess(
#         torch.FloatTensor(image_tensor.detach()),  # type: ignore
#         output_type="pil",
#         do_denormalize=[True],
#     )
#     image = image[0]  # type: ignore
#     assert isinstance(image, PIL.Image.Image)
#     return image



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class ModNet(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='input image')
    @classmethod
    def get_node_type(cls): return "replicate.image.process.ModNet"



class RemoveBackground(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    @classmethod
    def get_node_type(cls): return "replicate.image.process.RemoveBackground"





from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum


class RemoveBackground(ReplicateNode):
    """Remove images background"""

    @classmethod
    def replicate_model_id(cls):
        return "cjwbw/rembg:fb8af171cfa1616ddcf1242c093f9c46bcada5ad4cf6f2fbe8b81b330ec5c003"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/pbxt/2hczaMwD9xrsIR8h3Cl8iYGbHaCdFhIOMZ0LfoYfHlKuuIBQA/out.png",
            "created_at": "2022-11-18T00:55:22.939155Z",
            "description": "Remove images background",
            "github_url": "https://github.com/chenxwh/rembg/tree/replicate",
            "license_url": "https://github.com/danielgatis/rembg/blob/main/LICENSE.txt",
            "name": "rembg",
            "owner": "cjwbw",
            "paper_url": None,
            "run_count": 5271846,
            "url": "https://replicate.com/cjwbw/rembg",
            "visibility": "public",
            "hardware": "Nvidia A40 GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    image: ImageRef = Field(default=ImageRef(), description="Input image")


class ModNet(ReplicateNode):
    """A deep learning approach to remove background & adding new background image"""

    @classmethod
    def replicate_model_id(cls):
        return "pollinations/modnet:da7d45f3b836795f945f221fc0b01a6d3ab7f5e163f13208948ad436001e2255"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/bb0ab3e4-5efa-446f-939a-23e78f2b82de/output.png",
            "created_at": "2022-11-19T04:56:59.860128Z",
            "description": "A deep learning approach to remove background & adding new background image",
            "github_url": "https://github.com/pollinations/MODNet-BGRemover",
            "license_url": None,
            "name": "modnet",
            "owner": "pollinations",
            "paper_url": "https://arxiv.org/pdf/2011.11961.pdf",
            "run_count": 518632,
            "url": "https://replicate.com/pollinations/modnet",
            "visibility": "public",
            "hardware": "Nvidia T4 GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    image: ImageRef = Field(default=ImageRef(), description="input image")



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class HTTPDelete(GraphNode):
    url: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The URL to make the DELETE request to.')
    @classmethod
    def get_node_type(cls): return "nodetool.http.HTTPDelete"



class HTTPGet(GraphNode):
    url: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The URL to make the GET request to.')
    @classmethod
    def get_node_type(cls): return "nodetool.http.HTTPGet"



class HTTPPost(GraphNode):
    url: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The URL to make the POST request to.')
    data: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The data to send in the POST request.')
    @classmethod
    def get_node_type(cls): return "nodetool.http.HTTPPost"



class HTTPPut(GraphNode):
    url: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The URL to make the PUT request to.')
    data: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The data to send in the PUT request.')
    @classmethod
    def get_node_type(cls): return "nodetool.http.HTTPPut"





# Openpose
# Original from CMU https://github.com/CMU-Perceptual-Computing-Lab/openpose
# 2nd Edited by https://github.com/Hzzone/pytorch-openpose
# 3rd Edited by ControlNet
# 4th Edited by ControlNet (added face and correct hands)
# 5th Edited by ControlNet (Improved JSON serialization/deserialization, and lots of bug fixs)
# This preprocessor is licensed by CMU for non-commercial use only.


import os

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

import json
import warnings
from typing import Callable, List, NamedTuple, Tuple, Union

import cv2
import numpy as np
import torch
from huggingface_hub import hf_hub_download
from PIL import Image

from controlnet_aux.util import HWC3, common_input_validate, resize_image_with_pad, custom_hf_download, HF_MODEL_NAME
from . import util
from .body import Body, BodyResult, Keypoint
from .face import Face
from .hand import Hand

HandResult = List[Keypoint]
FaceResult = List[Keypoint]

class PoseResult(NamedTuple):
    body: BodyResult
    left_hand: Union[HandResult, None]
    right_hand: Union[HandResult, None]
    face: Union[FaceResult, None]

def draw_poses(poses: List[PoseResult], H, W, draw_body=True, draw_hand=True, draw_face=True):
    """
    Draw the detected poses on an empty canvas.

    Args:
        poses (List[PoseResult]): A list of PoseResult objects containing the detected poses.
        H (int): The height of the canvas.
        W (int): The width of the canvas.
        draw_body (bool, optional): Whether to draw body keypoints. Defaults to True.
        draw_hand (bool, optional): Whether to draw hand keypoints. Defaults to True.
        draw_face (bool, optional): Whether to draw face keypoints. Defaults to True.

    Returns:
        numpy.ndarray: A 3D numpy array representing the canvas with the drawn poses.
    """
    canvas = np.zeros(shape=(H, W, 3), dtype=np.uint8)

    for pose in poses:
        if draw_body:
            canvas = util.draw_bodypose(canvas, pose.body.keypoints)

        if draw_hand:
            canvas = util.draw_handpose(canvas, pose.left_hand)
            canvas = util.draw_handpose(canvas, pose.right_hand)

        if draw_face:
            canvas = util.draw_facepose(canvas, pose.face)

    return canvas

def encode_poses_as_json(poses: List[PoseResult], canvas_height: int, canvas_width: int) -> str:
    """ Encode the pose as a JSON string following openpose JSON output format:
    https://github.com/CMU-Perceptual-Computing-Lab/openpose/blob/master/doc/02_output.md
    """
    def compress_keypoints(keypoints: Union[List[Keypoint], None]) -> Union[List[float], None]:
        if not keypoints:
            return None
        
        return [
            value
            for keypoint in keypoints
            for value in (
                [float(keypoint.x), float(keypoint.y), 1.0]
                if keypoint is not None
                else [0.0, 0.0, 0.0]
            )
        ]

    return json.dumps({
        'people': [
            {
                'pose_keypoints_2d': compress_keypoints(pose.body.keypoints),
                "face_keypoints_2d": compress_keypoints(pose.face),
                "hand_left_keypoints_2d": compress_keypoints(pose.left_hand),
                "hand_right_keypoints_2d":compress_keypoints(pose.right_hand),
            }
            for pose in poses
        ],
        'canvas_height': canvas_height,
        'canvas_width': canvas_width,
    }, indent=4)
    
class OpenposeDetector:
    """
    A class for detecting human poses in images using the Openpose model.

    Attributes:
        model_dir (str): Path to the directory where the pose models are stored.
    """
    def __init__(self, body_estimation, hand_estimation=None, face_estimation=None):
        self.body_estimation = body_estimation
        self.hand_estimation = hand_estimation
        self.face_estimation = face_estimation

    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, filename="body_pose_model.pth", hand_filename="hand_pose_model.pth", face_filename="facenet.pth"):
        if pretrained_model_or_path == "lllyasviel/ControlNet":
            subfolder = "annotator/ckpts"
            face_pretrained_model_or_path = "lllyasviel/Annotators"
            
        else:
            subfolder = ''
            face_pretrained_model_or_path = pretrained_model_or_path

        body_model_path = custom_hf_download(pretrained_model_or_path, filename, subfolder=subfolder)
        hand_model_path = custom_hf_download(pretrained_model_or_path, hand_filename, subfolder=subfolder)
        face_model_path = custom_hf_download(face_pretrained_model_or_path, face_filename, subfolder=subfolder)

        body_estimation = Body(body_model_path)
        hand_estimation = Hand(hand_model_path)
        face_estimation = Face(face_model_path)

        return cls(body_estimation, hand_estimation, face_estimation)

    def to(self, device):
        self.body_estimation.to(device)
        self.hand_estimation.to(device)
        self.face_estimation.to(device)
        return self

    def detect_hands(self, body: BodyResult, oriImg) -> Tuple[Union[HandResult, None], Union[HandResult, None]]:
        left_hand = None
        right_hand = None
        H, W, _ = oriImg.shape
        for x, y, w, is_left in util.handDetect(body, oriImg):
            peaks = self.hand_estimation(oriImg[y:y+w, x:x+w, :]).astype(np.float32)
            if peaks.ndim == 2 and peaks.shape[1] == 2:
                peaks[:, 0] = np.where(peaks[:, 0] < 1e-6, -1, peaks[:, 0] + x) / float(W)
                peaks[:, 1] = np.where(peaks[:, 1] < 1e-6, -1, peaks[:, 1] + y) / float(H)
                
                hand_result = [
                    Keypoint(x=peak[0], y=peak[1])
                    for peak in peaks
                ]

                if is_left:
                    left_hand = hand_result
                else:
                    right_hand = hand_result

        return left_hand, right_hand

    def detect_face(self, body: BodyResult, oriImg) -> Union[FaceResult, None]:
        face = util.faceDetect(body, oriImg)
        if face is None:
            return None
        
        x, y, w = face
        H, W, _ = oriImg.shape
        heatmaps = self.face_estimation(oriImg[y:y+w, x:x+w, :])
        peaks = self.face_estimation.compute_peaks_from_heatmaps(heatmaps).astype(np.float32)
        if peaks.ndim == 2 and peaks.shape[1] == 2:
            peaks[:, 0] = np.where(peaks[:, 0] < 1e-6, -1, peaks[:, 0] + x) / float(W)
            peaks[:, 1] = np.where(peaks[:, 1] < 1e-6, -1, peaks[:, 1] + y) / float(H)
            return [
                Keypoint(x=peak[0], y=peak[1])
                for peak in peaks
            ]
        
        return None

    def detect_poses(self, oriImg, include_hand=False, include_face=False) -> List[PoseResult]:
        """
        Detect poses in the given image.
            Args:
                oriImg (numpy.ndarray): The input image for pose detection.
                include_hand (bool, optional): Whether to include hand detection. Defaults to False.
                include_face (bool, optional): Whether to include face detection. Defaults to False.

        Returns:
            List[PoseResult]: A list of PoseResult objects containing the detected poses.
        """
        oriImg = oriImg[:, :, ::-1].copy()
        H, W, C = oriImg.shape
        with torch.no_grad():
            candidate, subset = self.body_estimation(oriImg)
            bodies = self.body_estimation.format_body_result(candidate, subset)

            results = []
            for body in bodies:
                left_hand, right_hand, face = (None,) * 3
                if include_hand:
                    left_hand, right_hand = self.detect_hands(body, oriImg)
                if include_face:
                    face = self.detect_face(body, oriImg)
                
                results.append(PoseResult(BodyResult(
                    keypoints=[
                        Keypoint(
                            x=keypoint.x / float(W),
                            y=keypoint.y / float(H)
                        ) if keypoint is not None else None
                        for keypoint in body.keypoints
                    ], 
                    total_score=body.total_score,
                    total_parts=body.total_parts
                ), left_hand, right_hand, face))
            
            return results
        
    def __call__(self, input_image, detect_resolution=512, include_body=True, include_hand=False, include_face=False, hand_and_face=None, output_type="pil", image_and_json=False, upscale_method="INTER_CUBIC", **kwargs):
        if hand_and_face is not None:
            warnings.warn("hand_and_face is deprecated. Use include_hand and include_face instead.", DeprecationWarning)
            include_hand = hand_and_face
            include_face = hand_and_face

        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)
        
        poses = self.detect_poses(input_image, include_hand=include_hand, include_face=include_face)
        canvas = draw_poses(poses, input_image.shape[0], input_image.shape[1], draw_body=include_body, draw_hand=include_hand, draw_face=include_face) 
        detected_map = HWC3(remove_pad(canvas))

        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
        
        if image_and_json:
            return (detected_map, encode_poses_as_json(poses, detected_map.shape[0], detected_map.shape[1]))
        
        return detected_map



import os
import warnings

import cv2
import numpy as np
import torch
from PIL import Image

from controlnet_aux.util import HWC3, common_input_validate, resize_image_with_pad, custom_hf_download, HF_MODEL_NAME
from .models.mbv2_mlsd_large import MobileV2_MLSD_Large
from .utils import pred_lines


class MLSDdetector:
    def __init__(self, model):
        self.model = model

    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, filename="mlsd_large_512_fp32.pth"):
        subfolder = "annotator/ckpts" if pretrained_model_or_path == "lllyasviel/ControlNet" else ''
        model_path = custom_hf_download(pretrained_model_or_path, filename, subfolder=subfolder)
        model = MobileV2_MLSD_Large()
        model.load_state_dict(torch.load(model_path), strict=True)
        model.eval()

        return cls(model)

    def to(self, device):
        self.model.to(device)
        return self
    
    def __call__(self, input_image, thr_v=0.1, thr_d=0.1, detect_resolution=512, output_type="pil", upscale_method="INTER_AREA", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        detected_map, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)
        img = detected_map
        img_output = np.zeros_like(img)
        try:
            with torch.no_grad():
                lines = pred_lines(img, self.model, [img.shape[0], img.shape[1]], thr_v, thr_d)
                for line in lines:
                    x_start, y_start, x_end, y_end = [int(val) for val in line]
                    cv2.line(img_output, (x_start, y_start), (x_end, y_end), [255, 255, 255], 1)
        except Exception as e:
            pass

        detected_map = remove_pad(HWC3(img_output[:, :, 0]))

        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map



import custom_timm as timm

from .swin_common import _make_swin_backbone


def _make_pretrained_swinl12_384(pretrained, hooks=None):
    model = timm.create_model("swin_large_patch4_window12_384", pretrained=pretrained)

    hooks = [1, 1, 17, 1] if hooks == None else hooks
    return _make_swin_backbone(
        model,
        hooks=hooks
    )



from typing import Any
from pydantic import BaseModel, Field

from nodetool.common.environment import Environment
from uuid import uuid1
from random import randint


log = Environment.get_logger()


def create_time_ordered_uuid() -> str:
    """
    Create an uuid that is ordered by time.
    """
    return uuid1(randint(0, 2**31)).hex


def DBField(hash_key: bool = False, **kwargs: Any):
    return Field(json_schema_extra={"hash_key": hash_key, "persist": True}, **kwargs)  # type: ignore


class DBModel(BaseModel):
    @classmethod
    def get_table_schema(cls) -> dict[str, Any]:
        """
        Get the name of the table for the model.
        """
        raise NotImplementedError()

    @classmethod
    def get_table_name(cls) -> str:
        """
        Get the name of the table for the model.
        """
        return cls.get_table_schema()["table_name"]

    @classmethod
    def adapter(cls):
        if not hasattr(cls, "__adapter"):
            cls.__adapter = Environment.get_database_adapter(
                fields=cls.db_fields(),
                table_schema=cls.get_table_schema(),
            )
        return cls.__adapter

    @classmethod
    def create_table(cls):
        """
        Create the DB table for the model.
        """
        cls.adapter().create_table()

    @classmethod
    def drop_table(cls):
        """
        Drop the DB table for the model.
        """
        cls.adapter().drop_table()

    @classmethod
    def query(
        cls,
        condition: str,
        values: dict[str, Any],
        index: str | None = None,
        start_key: str | None = None,
        limit: int = 100,
        reverse: bool = False,
    ):
        """
        Query the DB table for the model to retrieve a list of items.
        This method is used for pagination and returns a tuple containing a list of items and the last evaluated key.
        It allows for filtering and sorting the results.

        Args:
            condition: The condition for the query based on the primary key.
            values: The values for the condition.
            index: The name of the index to use for the query.
            start_key: The exclusive start key for pagination.
            limit: The maximum number of items to retrieve.
            reverse: Whether to reverse the order of the results.

        Returns:
            A tuple containing a list of items that match the query conditions and the last evaluated key.
        """
        items, key = cls.adapter().query(
            condition=condition,
            values=values,
            index=index,
            start_key=start_key,
            limit=limit,
            reverse=reverse,
        )
        return [cls(**item) for item in items], key

    @classmethod
    def create(cls, **kwargs):
        """
        Create a model instance from keyword arguments and save it to DynamoDB.
        """
        return cls(**kwargs).save()

    def before_save(self):
        pass

    def save(self):
        """
        Save a model instance to DynamoDB and return the instance.
        """
        self.before_save()
        self.adapter().save(self.model_dump())
        return self

    @classmethod
    def db_fields(cls) -> dict[str, Any]:
        """
        Return a dictionary of fields that should be persisted to DynamoDB.
        """
        return {
            field_name: field
            for field_name, field in cls.model_fields.items()
            if field.json_schema_extra and field.json_schema_extra.get("persist", False)
        }

    @classmethod
    def get(cls, key: str | int):
        """
        Retrieve a model instance from the DB using a key.
        """
        item = cls.adapter().get(key)
        if item is None:
            return None
        return cls(**item)

    def partition_value(self) -> str:
        """
        Get the value of the hash key.
        """
        return getattr(self, self.adapter().get_primary_key())

    def delete(self):
        self.adapter().delete(self.partition_value())

    def update(self, **kwargs):
        """
        Update the model instance and save it to DynamoDB.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self.save()



from transformers import CLIPSegProcessor, CLIPSegForImageSegmentation

from PIL import Image
import torch
import torchvision.transforms as T
import numpy as np

from torchvision.transforms.functional import to_pil_image
import matplotlib.pyplot as plt
import matplotlib.cm as cm


import cv2

from scipy.ndimage import gaussian_filter

from typing import Optional, Tuple

import warnings
warnings.filterwarnings("ignore", category=UserWarning, module="torch")
warnings.filterwarnings("ignore", category=UserWarning, module="safetensors")



"""Helper methods for CLIPSeg nodes"""

def tensor_to_numpy(tensor: torch.Tensor) -> np.ndarray:
    """Convert a tensor to a numpy array and scale its values to 0-255."""
    array = tensor.numpy().squeeze()
    return (array * 255).astype(np.uint8)

def numpy_to_tensor(array: np.ndarray) -> torch.Tensor:
    """Convert a numpy array to a tensor and scale its values from 0-255 to 0-1."""
    array = array.astype(np.float32) / 255.0
    return torch.from_numpy(array)[None,]

def apply_colormap(mask: torch.Tensor, colormap) -> np.ndarray:
    """Apply a colormap to a tensor and convert it to a numpy array."""
    colored_mask = colormap(mask.numpy())[:, :, :3]
    return (colored_mask * 255).astype(np.uint8)

def resize_image(image: np.ndarray, dimensions: Tuple[int, int]) -> np.ndarray:
    """Resize an image to the given dimensions using linear interpolation."""
    return cv2.resize(image, dimensions, interpolation=cv2.INTER_LINEAR)

def overlay_image(background: np.ndarray, foreground: np.ndarray, alpha: float) -> np.ndarray:
    """Overlay the foreground image onto the background with a given opacity (alpha)."""
    return cv2.addWeighted(background, 1 - alpha, foreground, alpha, 0)

def dilate_mask(mask: torch.Tensor, dilation_factor: float) -> torch.Tensor:
    """Dilate a mask using a square kernel with a given dilation factor."""
    kernel_size = int(dilation_factor * 2) + 1
    kernel = np.ones((kernel_size, kernel_size), np.uint8)
    mask_dilated = cv2.dilate(mask.numpy(), kernel, iterations=1)
    return torch.from_numpy(mask_dilated)



class CLIPSeg:

    def __init__(self):
        pass
    
    @classmethod
    def INPUT_TYPES(s):
        """
            Return a dictionary which contains config for all input fields.
            Some types (string): "MODEL", "VAE", "CLIP", "CONDITIONING", "LATENT", "IMAGE", "INT", "STRING", "FLOAT".
            Input types "INT", "STRING" or "FLOAT" are special values for fields on the node.
            The type can be a list for selection.

            Returns: `dict`:
                - Key input_fields_group (`string`): Can be either required, hidden or optional. A node class must have property `required`
                - Value input_fields (`dict`): Contains input fields config:
                    * Key field_name (`string`): Name of a entry-point method's argument
                    * Value field_config (`tuple`):
                        + First value is a string indicate the type of field or a list for selection.
                        + Secound value is a config for type "INT", "STRING" or "FLOAT".
        """
        return {"required":
                    {
                        "image": ("IMAGE",),
                        "text": ("STRING", {"multiline": False}),
                        
                     },
                "optional":
                    {
                        "blur": ("FLOAT", {"min": 0, "max": 15, "step": 0.1, "default": 7}),
                        "threshold": ("FLOAT", {"min": 0, "max": 1, "step": 0.05, "default": 0.4}),
                        "dilation_factor": ("INT", {"min": 0, "max": 10, "step": 1, "default": 4}),
                    }
                }

    CATEGORY = "image"
    RETURN_TYPES = ("MASK", "IMAGE", "IMAGE",)
    RETURN_NAMES = ("Mask","Heatmap Mask", "BW Mask")

    FUNCTION = "segment_image"
    def segment_image(self, image: torch.Tensor, text: str, blur: float, threshold: float, dilation_factor: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Create a segmentation mask from an image and a text prompt using CLIPSeg.

        Args:
            image (torch.Tensor): The image to segment.
            text (str): The text prompt to use for segmentation.
            blur (float): How much to blur the segmentation mask.
            threshold (float): The threshold to use for binarizing the segmentation mask.
            dilation_factor (int): How much to dilate the segmentation mask.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: The segmentation mask, the heatmap mask, and the binarized mask.
        """
            
        # Convert the Tensor to a PIL image
        image_np = image.numpy().squeeze()  # Remove the first dimension (batch size of 1)
        # Convert the numpy array back to the original range (0-255) and data type (uint8)
        image_np = (image_np * 255).astype(np.uint8)
        # Create a PIL image from the numpy array
        i = Image.fromarray(image_np, mode="RGB")

        processor = CLIPSegProcessor.from_pretrained("CIDAS/clipseg-rd64-refined")
        model = CLIPSegForImageSegmentation.from_pretrained("CIDAS/clipseg-rd64-refined")
        
        prompt = text
        
        input_prc = processor(text=prompt, images=i, padding="max_length", return_tensors="pt")
        
        # Predict the segemntation mask
        with torch.no_grad():
            outputs = model(**input_prc)
        
        tensor = torch.sigmoid(outputs[0]) # get the mask
        
        # Apply a threshold to the original tensor to cut off low values
        thresh = threshold
        tensor_thresholded = torch.where(tensor > thresh, tensor, torch.tensor(0, dtype=torch.float))

        # Apply Gaussian blur to the thresholded tensor
        sigma = blur
        tensor_smoothed = gaussian_filter(tensor_thresholded.numpy(), sigma=sigma)
        tensor_smoothed = torch.from_numpy(tensor_smoothed)

        # Normalize the smoothed tensor to [0, 1]
        mask_normalized = (tensor_smoothed - tensor_smoothed.min()) / (tensor_smoothed.max() - tensor_smoothed.min())

        # Dilate the normalized mask
        mask_dilated = dilate_mask(mask_normalized, dilation_factor)

        # Convert the mask to a heatmap and a binary mask
        heatmap = apply_colormap(mask_dilated, cm.viridis)
        binary_mask = apply_colormap(mask_dilated, cm.Greys_r)

        # Overlay the heatmap and binary mask on the original image
        dimensions = (image_np.shape[1], image_np.shape[0])
        heatmap_resized = resize_image(heatmap, dimensions)
        binary_mask_resized = resize_image(binary_mask, dimensions)

        alpha_heatmap, alpha_binary = 0.5, 1
        overlay_heatmap = overlay_image(image_np, heatmap_resized, alpha_heatmap)
        overlay_binary = overlay_image(image_np, binary_mask_resized, alpha_binary)

        # Convert the numpy arrays to tensors
        image_out_heatmap = numpy_to_tensor(overlay_heatmap)
        image_out_binary = numpy_to_tensor(overlay_binary)

        # Save or display the resulting binary mask
        binary_mask_image = Image.fromarray(binary_mask_resized[..., 0])

        # convert PIL image to numpy array
        tensor_bw = binary_mask_image.convert("RGB")
        tensor_bw = np.array(tensor_bw).astype(np.float32) / 255.0
        tensor_bw = torch.from_numpy(tensor_bw)[None,]
        tensor_bw = tensor_bw.squeeze(0)[..., 0]

        return tensor_bw, image_out_heatmap, image_out_binary

    #OUTPUT_NODE = False

class CombineMasks:
    def __init__(self):
        pass

    @classmethod
    def INPUT_TYPES(s):
        return {"required":
                    {
                        "input_image": ("IMAGE", ),
                        "mask_1": ("MASK", ), 
                        "mask_2": ("MASK", ),
                    },
                "optional": 
                    {
                        "mask_3": ("MASK",), 
                    },
                }
        
    CATEGORY = "image"
    RETURN_TYPES = ("MASK", "IMAGE", "IMAGE",)
    RETURN_NAMES = ("Combined Mask","Heatmap Mask", "BW Mask")

    FUNCTION = "combine_masks"
            
    def combine_masks(self, input_image: torch.Tensor, mask_1: torch.Tensor, mask_2: torch.Tensor, mask_3: Optional[torch.Tensor] = None) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """A method that combines two or three masks into one mask. Takes in tensors and returns the mask as a tensor, as well as the heatmap and binary mask as tensors."""

        # Combine masks
        combined_mask = mask_1 + mask_2 + mask_3 if mask_3 is not None else mask_1 + mask_2


        # Convert image and masks to numpy arrays
        image_np = tensor_to_numpy(input_image)
        heatmap = apply_colormap(combined_mask, cm.viridis)
        binary_mask = apply_colormap(combined_mask, cm.Greys_r)

        # Resize heatmap and binary mask to match the original image dimensions
        dimensions = (image_np.shape[1], image_np.shape[0])
        heatmap_resized = resize_image(heatmap, dimensions)
        binary_mask_resized = resize_image(binary_mask, dimensions)

        # Overlay the heatmap and binary mask onto the original image
        alpha_heatmap, alpha_binary = 0.5, 1
        overlay_heatmap = overlay_image(image_np, heatmap_resized, alpha_heatmap)
        overlay_binary = overlay_image(image_np, binary_mask_resized, alpha_binary)

        # Convert overlays to tensors
        image_out_heatmap = numpy_to_tensor(overlay_heatmap)
        image_out_binary = numpy_to_tensor(overlay_binary)

        return combined_mask, image_out_heatmap, image_out_binary

# A dictionary that contains all nodes you want to export with their names
# NOTE: names should be globally unique
NODE_CLASS_MAPPINGS = {
    "CLIPSeg": CLIPSeg,
    "CombineSegMasks": CombineMasks,
}



from pydantic import Field
from nodetool.metadata.types import Conditioning, ControlNet, ImageTensor
from nodetool.common.comfy_node import MAX_RESOLUTION
from nodetool.common.comfy_node import ComfyNode


class PreprocessImage(ComfyNode):
    image: ImageTensor = Field(
        default=ImageTensor(), description="The image to preprocess."
    )
    resolution: int = Field(
        default=512,
        description="The width of the image to generate.",
        ge=64,
        le=MAX_RESOLUTION,
    )

    @classmethod
    def return_type(cls):
        return {"image": ImageTensor}



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class Append(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    value: Any | GraphNode | tuple[GraphNode, str] = Field(default=None, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.Append"



class Dedupe(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.Dedupe"



class Extend(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    other_values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.Extend"



class Filter(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='The list to filter.')
    condition: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The Python code to use as the filtering condition.')
    @classmethod
    def get_node_type(cls): return "nodetool.list.Filter"



class Index(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    index: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.Index"



class Length(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.Length"



class Map(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='The list to map.')
    code: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Python code to use for mapping.')
    @classmethod
    def get_node_type(cls): return "nodetool.list.Map"



class Range(GraphNode):
    start: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    stop: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    step: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.Range"



class Reduce(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='The list to reduce.')
    initial_value: Any | GraphNode | tuple[GraphNode, str] = Field(default=None, description='The initial value for the reduction. If not provided, the first value in the list is used.')
    reduction_code: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The Python code to use for the reduction.')
    @classmethod
    def get_node_type(cls): return "nodetool.list.Reduce"



class Reverse(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.Reverse"



class SaveList(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='The list to save.')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='text.txt', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.SaveList"



class Select(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    indices: list[int] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.Select"



class Slice(GraphNode):
    values: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    start: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    stop: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    step: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.list.Slice"





import asyncio
from typing import Any
from pydantic import Field
from nodetool.metadata.types import ImageRef
from nodetool.common.huggingface_node import HuggingfaceNode
from nodetool.workflows.processing_context import ProcessingContext


class Segformer(HuggingfaceNode):
    image: ImageRef = Field(
        default=ImageRef(),
        title="Image",
        description="The input image to segment",
    )

    async def process(self, context: ProcessingContext) -> dict[str, ImageRef]:
        image = await context.asset_to_io(self.image)
        result = await self.run_huggingface(
            model_id="nvidia/segformer-b3-finetuned-ade-512-512",
            context=context,
            data=image.read(),
        )

        async def convert_output(item: dict[str, Any]):
            mask = await context.image_from_base64(item["mask"])
            return item["label"], mask

        items = await asyncio.gather(*[convert_output(item) for item in list(result)])
        return {label: mask for label, mask in items}



import math
import numpy as np
from typing import NamedTuple

import PIL.Image


class Tile(NamedTuple):
    image: PIL.Image.Image
    x: int
    y: int


def make_grid(
    w: int, h: int, tile_w=512, tile_h=512, overlap=64
) -> tuple[list[list[tuple[int, int]]], int, int]:
    """
    Make a grid of tiles with the given width and height. The tiles will be
    overlapping, with the given overlap size.
    """
    assert tile_h <= h, "tile_h must be less than or equal to h"
    assert overlap < tile_w, "overlap must be less than tile_w"
    assert overlap < tile_h, "overlap must be less than tile_h"
    assert w > 0, "w must be greater than 0"
    assert h > 0, "h must be greater than 0"
    assert overlap >= 0, "overlap must be greater than or equal to 0"

    # The width and height of the non-overlapping part of the tile
    non_overlap_width = tile_w - overlap
    non_overlap_height = tile_h - overlap

    # The number of rows and columns of tiles
    cols = math.ceil((w - overlap) / non_overlap_width)
    rows = math.ceil((h - overlap) / non_overlap_height)

    # The distance between the left edge of each tile
    # If there is only one column, dx will be 0
    # If there are multiple columns, dx will be the distance between the left edge of each tile
    dx = (w - tile_w) / (cols - 1) if cols > 1 else 0
    dy = (h - tile_h) / (rows - 1) if rows > 1 else 0

    tiles = []

    for row in range(rows):
        row_tiles = []
        y = int(row * dy)

        # Align the bottom edge of the last tile with the bottom edge of the image
        if y + tile_h >= h:
            y = h - tile_h

        for col in range(cols):
            x = int(col * dx)

            # Align the right edge of the last tile with the right edge of the image
            if x + tile_w >= w:
                x = w - tile_w

            row_tiles.append((x, y))

        tiles.append(row_tiles)

    return tiles, cols, rows


def in_groups_of(items: list, group_size: int, fill_value=None) -> list[list[any]]:
    """
    Split a list into groups of the given size.

    The last group may be smaller than the given size.

    Args:
        items: A list of items to split into groups.
        group_size: The size of each group.
        fill_value: The value to use to fill the last group if it is smaller than the given size.
    """
    groups = []
    for i in range(0, len(items), group_size):
        group = items[i : i + group_size]
        if len(group) < group_size:
            group = group + [fill_value] * (group_size - len(group))
        groups.append(group)
    return groups


def flatten(items: list[list[any]]) -> list[any]:
    """
    Flatten a list of lists into a single list.
    """
    return [item for sublist in items for item in sublist]


def combine_grid(
    tiles: list[list[Tile]],
    tile_w: int,
    tile_h: int,
    image_w: int,
    image_h: int,
    overlap: int,
):
    """
    Combine a grid of tiles into a single image.

    The tiles will be tile_w x tile_h pixels, with overlap pixels of overlap.

    The tiles must be in the same order as the grid returned by make_grid.

    The tiles must be overlapping, with the given overlap size.

    Args:
        tiles: A list of lists of tiles.
        tile_w: The width of the tiles.
        tile_h: The height of the tiles.
        image_w: The width of the image.
        image_h: The height of the image.
        overlap: The number of pixels of overlap between tiles.
    """

    def make_mask_image(r):
        """
        Make a mask image from a 2D array of values.

        The values will be scaled to 0-255.
        """
        r = r.astype(np.uint8)
        return PIL.Image.fromarray(r, "L")

    # Mask image for the left edge of the tile
    # The size of the mask image is (overlap, tile_h)
    # The values of the mask image are 0-255, with 0 being transparent and 255 being opaque.
    mask_w = make_mask_image(
        np.arange(overlap, dtype=np.float32)
        .reshape((1, overlap))
        .repeat(tile_h, axis=0)
    )

    # Mask image for the top edge of the tile
    # The size of the mask image is (tile_w, overlap)
    # The values of the mask image are 0-255, with 0 being transparent and 255 being opaque.
    mask_h = make_mask_image(
        np.arange(overlap, dtype=np.float32)
        .reshape((overlap, 1))
        .repeat(image_w, axis=1)
    )

    combined_image = PIL.Image.new("RGB", (image_w, image_h))

    # Combine the tiles into rows and columns
    for row in tiles:
        combined_row = PIL.Image.new("RGB", (image_w, tile_h))
        for tile in row:
            if tile.x == 0:
                # The first tile in the row is not overlapping with the previous tile
                # Just paste the tile into the row and continue
                # This is a special case because the tile does not need to be combined with the previous tile
                combined_row.paste(tile.image, (0, 0))
                continue

            # Combine the tile with the left and top edges of the previous tile
            # The left edge of the tile is overlapping with the right edge of the previous tile.
            # The top edge of the tile is overlapping with the bottom edge of the previous tile.
            # The mask image is used to make the overlapping parts of the tile transparent.
            combined_row.paste(
                tile.image.crop((0, 0, overlap, tile_h)), (tile.x, 0), mask=mask_w
            )

            # The remaining part of the tile is not overlapping with the previous tile
            # Just paste the tile into the row.
            combined_row.paste(
                tile.image.crop((overlap, 0, tile_w, tile_h)), (tile.x + overlap, 0)
            )

        if tile.y == 0:
            # The first row is not overlapping with the previous row
            # Just paste the row into the combined image and continue
            combined_image.paste(combined_row, (0, 0))
            continue

        # Combine the row with the left and top edges of the previous row
        # The top edge of the row is overlapping with the bottom edge of the previous row.
        # The mask image is used to make the overlapping parts of the row transparent.
        combined_image.paste(
            combined_row.crop((0, 0, combined_row.width, overlap)),
            (0, tile.y),
            mask=mask_h,
        )
        # The remaining part of the row is not overlapping with the previous row
        # Just paste the row into the combined image.
        combined_image.paste(
            combined_row.crop((0, overlap, combined_row.width, tile_h)),
            (0, tile.y + overlap),
        )

    return combined_image



import PIL.Image
import PIL.ImageOps
import PIL.ImageFilter
import PIL.ImageEnhance
import numpy as np

from nodetool.metadata.types import Field, ImageRef
from nodetool.workflows.base_node import BaseNode
from nodetool.workflows.processing_context import ProcessingContext
from pydantic import Field


def adaptive_contrast(
    image: PIL.Image.Image, clip_limit: float, grid_size: int
) -> PIL.Image.Image:
    import cv2

    img = np.array(image)

    # Convert image from BGR to LAB color model
    img_lab = cv2.cvtColor(img, cv2.COLOR_BGR2Lab)

    # Split the LAB image into L, A and B channels
    l, a, b = cv2.split(img_lab)

    # Perform histogram equalization only on the L channel
    clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=(grid_size, grid_size))
    cl = clahe.apply(l)

    # Merge the CLAHE enhanced L channel with the original A and B channel
    merged_channels = cv2.merge((cl, a, b))

    # Convert image from LAB color model back to BGR
    final_img = cv2.cvtColor(merged_channels, cv2.COLOR_Lab2BGR)

    return PIL.Image.fromarray(final_img)


def sharpen_image(image: PIL.Image.Image) -> PIL.Image.Image:
    # Convert PIL image to CV
    import cv2

    img = np.array(image)
    kernel = np.array([[-1, -1, -1], [-1, 9, -1], [-1, -1, -1]])
    img_enhanced = cv2.filter2D(img, -1, kernel)
    return PIL.Image.fromarray(img_enhanced)


def canny_edge_detection(
    image: PIL.Image.Image, low_threshold: int, high_threshold: int
) -> PIL.Image.Image:
    import cv2

    arr = np.array(image)
    arr = cv2.Canny(arr, low_threshold, high_threshold)  # type: ignore
    arr = arr[:, :, None]
    arr = np.concatenate([arr, arr, arr], axis=2)
    return PIL.Image.fromarray(arr)


class AutoContrast(BaseNode):
    """
    This node automatically adjusts the contrast of an image.

    The purpose of the Auto Contrast Node is to enhance the quality of an image by optimizing its contrast levels. It intelligently adjusts the contrast of an image so brighter areas are brightened further and darker areas darkened, balancing the image illumination. The degree of contrast adjustment can be configured using a cutoff setting.

    #### Applications
    - Image processing: Enhance the visual quality of images by balancing the contrast.
    - Pre-processing in computer vision: Prepare images for further analysis by improving their readability.
    - Photo editing: Adjust the contrast in images for aesthetic or clarity purposes.
    """

    image: ImageRef = Field(
        default=ImageRef(), description="The image to adjust the contrast for."
    )
    cutoff: int = Field(default=0, ge=0, le=255, description="Cutoff for autocontrast.")

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        img = PIL.ImageOps.autocontrast(image, cutoff=self.cutoff)
        return await context.image_from_pil(img)


class Sharpness(BaseNode):
    """
    The Sharpness Node is used to apply a sharpness filter to images.

    The Sharpness Node adjusts the sharpness level of an image, allowing you to enhance or reduce its sharpness based on a factor you provide. A sharpness factor of 1.0 implies no change to the image's sharpness level.

    #### Applications
    - Photo editing: Enhance the sharpness of a particular photo to make it more visually appealing.
    - Image processing: It helps in refining images for purposes such as object detection, image recognition etc.
    """

    image: ImageRef = Field(
        default=ImageRef(), description="The image to adjust the brightness for."
    )
    factor: float = Field(
        default=1.0, description="Factor to adjust the contrast. 1.0 means no change."
    )

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        res = PIL.ImageEnhance.Sharpness(image).enhance(self.factor)
        return await context.image_from_pil(res)


class Equalize(BaseNode):
    """
    The Equalize Node is used to apply an equalizing filter to an image.

    Equalizing an image helps in enhancing its contrast by spreading out the most frequent intensity values throughout the image. The Equalize Node accomplishes this by transforming the values in pixel intensity histograms to create a uniform distribution of grayscale values.

    #### Applications
    - Improving the quality of images have poor lighting conditions.
    - Enhancing the details of an image for better viewing or analysis.
    """

    image: ImageRef = Field(default=ImageRef(), description="The image to equalize.")

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        res = PIL.ImageOps.equalize(image)
        return await context.image_from_pil(res)


class Contrast(BaseNode):
    """
    The Contrast Node is a simple tool that adjust the contrast of an image.

    The Contrast Node is used for modifying the distinction between the lightest and darkest areas of an image. It adjusts the contrast based on a provided factor. Setting the contrast factor to 1.0 will make no changes to the original image. Higher contrast makes the image more dynamic and lower contrast makes the image less dynamic.

    #### Applications
    - Altering the visibility of textures in an image.
    - Improving the clarity of an image for analysis or display.
    - Giving images a dramatic or subtle look according to preferences.
    """

    image: ImageRef = Field(
        default=ImageRef(), description="The image to adjust the brightness for."
    )
    factor: float = Field(
        default=1.0, description="Factor to adjust the contrast. 1.0 means no change."
    )

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        res = PIL.ImageEnhance.Contrast(image).enhance(self.factor)
        return await context.image_from_pil(res)


class EdgeEnhance(BaseNode):
    """
    This is a node that applies an edge enhance filter to an image.

    The Edge Enhance Node is used to improve the visibility of edges in an image. It is a type of image processing operation that enhances the edge contrast of an image, making the boundaries of objects within the image more pronounced. This node makes use of the edge enhance filter to perform its function.

    #### Applications
    - Image editing and enhancement: Enhancing the edges of an object in an image makes it easier to distinguish and recognize.
    - Object detection: Improving the edge contrast can make the process of object detection more accurate.
    - Digital art creation: Enhance edges of your artworks to add more contrast and depth.
    """

    image: ImageRef = Field(
        default=ImageRef(), description="The image to edge enhance."
    )

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        return await context.image_from_pil(image.filter(PIL.ImageFilter.EDGE_ENHANCE))


class Sharpen(BaseNode):
    """
    The Sharpen Node is a tool that applies a sharpening filter to an image.

    The primary purpose of this node is to enhance the visibility of details in an image. It performs this by intensifying the contrast of adjacent pixels, which provides a sharper and clearer image. It brings a new dimension of clarity, which is particularly useful in detecting subtle elements in images.

    #### Applications
    - Photo Editing: Enhancing the clarity of images to bring out small details.
    - Object Detection: To improve the detection accuracy by sharpening blurry images.
    """

    image: ImageRef = Field(default=ImageRef(), description="The image to sharpen.")

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        return await context.image_from_pil(image.filter(PIL.ImageFilter.SHARPEN))


class RankFilter(BaseNode):
    """
    The Rank Filter Node is used to apply a rank filter to a given image.

    The purpose of this node is to perform a specific type of image processing technique, known as a rank filter. A rank filter considers the surrounding pixels of each pixel in the image and sorts them according to brightness. The 'rank' refers to the position of the pixel that will replace the current pixel - for example, a rank of 1 would replace each pixel with the darkest surrounding pixel, while a rank of 3 would replace it with the third darkest and so on. The 'size' indicates the total number of surrounding pixels to be considered for this process.

    This node is pivotal for tasks related to image enhancement and noise reduction and plays a major role in improving the visual quality of an image, making it more suitable for further processing and analysis.

    #### Applications
    - Image enhancement: This node can be used to enhance the quality of an image, making subtle details more pronounced.
    - Noise reduction: The Rank Filter Node can help reduce 'noise' in images, i.e., unwanted or irrelevant details.
    """

    image: ImageRef = Field(default=ImageRef(), description="The image to rank filter.")
    size: int = Field(default=3, ge=1, le=512, description="Rank filter size.")
    rank: int = Field(default=3, ge=1, le=512, description="Rank filter rank.")

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        return await context.image_from_pil(
            image.filter(PIL.ImageFilter.RankFilter(self.size, self.rank))
        )


class UnsharpMask(BaseNode):
    """
    The Unsharp Mask Node is responsible for performing the unsharp mask filter on an image.

    An unsharp mask is a photographic sharpening technique. This node helps to enhance the sharpness of your image by adjusting the radius, the impact and the threshold of the unsharp mask filter.

    #### Applications
    - Photo editing: Sharpens images by emphasizing transitions, such as those defining edges of objects in the image.
    - Digital art creation: Refining and adding crispness to digital artwork.
    """

    image: ImageRef = Field(
        default=ImageRef(), description="The image to unsharp mask."
    )
    radius: int = Field(default=2, ge=0, le=512, description="Unsharp mask radius.")
    percent: int = Field(
        default=150, ge=0, le=1000, description="Unsharp mask percent."
    )
    threshold: int = Field(
        default=3, ge=0, le=512, description="Unsharp mask threshold."
    )

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        res = image.filter(
            PIL.ImageFilter.UnsharpMask(self.radius, self.percent, self.threshold)
        )
        return await context.image_from_pil(res)


class Brightness(BaseNode):
    """
    The Brightness Node is designed to adjust the brightness of an image.

    Its main purpose is to alter the brightness of an uploaded image by using a customizable factor. The brightness of an image can influence its clarity and visibility. By changing the brightness levels, we can make portions of an image more visible or less distracting based on our needs.

    #### Applications
    - Brightness level correction: If an image is too dark or too bright, the Brightness Node can be used to correct the brightness levels to the desired range.
    - Highlighting or dimming certain parts of an image: The Brightness Node can be used to increase the visibility of certain parts of an image by enhancing their brightness.
    """

    image: ImageRef = Field(
        default=ImageRef(), description="The image to adjust the brightness for."
    )
    factor: float | int = Field(
        default=1.0, description="Factor to adjust the brightness. 1.0 means no change."
    )

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        return await context.image_from_pil(
            PIL.ImageEnhance.Brightness(image).enhance(self.factor)
        )


class Color(BaseNode):
    """
    The ColorNode is a simple tool that applies a color filter to an image.

    It's an excellent option if you want to enhance or tweak the colors in your images for a specific visual effect. The node allows you to control the degree of color change through the factor input. A factor of 1.0 means no color change, while values greater or smaller than 1.0 increase or decrease the image's color intensity, respectively.

    #### Applications
    - Color grading: You can use it to adjust the colors in a photo for aesthetic purposes.
    - Highlighting features: Enhance colors to make certain areas of an image stand out.
    - Color correction: Correct or adjust the color balance in an image.
    """

    image: ImageRef = Field(
        default=ImageRef(), description="The image to adjust the brightness for."
    )
    factor: float = Field(
        default=1.0, description="Factor to adjust the contrast. 1.0 means no change."
    )

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        res = PIL.ImageEnhance.Color(image).enhance(self.factor)
        return await context.image_from_pil(res)


class Detail(BaseNode):
    """
    The DetailNode is an image detail filter.

    This node is primarily designed to enhance the details of an image in an AI workflow by applying a detail filter. It is especially useful when you want to emphasize or highlight certain features of the image.

    #### Applications
    - Enhancing the details of an image to improve its clarity.
    - Highlighting certain features of an image for a deeper analysis.
    """

    image: ImageRef = Field(default=ImageRef(), description="The image to detail.")

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        return await context.image_from_pil(image.filter(PIL.ImageFilter.DETAIL))


class AdaptiveContrast(BaseNode):
    """
    The Adaptive Contrast Node is used to adjust the contrasting visual aspects of an image.

    This node enhances the contrast of the provided image on a localized level by adapting the contrast based on defined grid sizes, which means each section of the image will have its contrast adjusted individually compared to each other. The balance of brightness and darkness in different various parts of the image can be shifted using the clip limit field.

    #### Applications
    - Enhancing visual interest in an image by adjusting contrast.
    - Mitigation of contrasting errors in color-focused or texture-focused AI models.
    - For image restoration, where restoring the contrast levels can make old or degraded photos become clearer.
    """

    image: ImageRef = Field(
        default=ImageRef(), description="The image to adjust the contrast for."
    )
    clip_limit: float = Field(
        default=2.0, ge=0.0, le=100, description="Clip limit for adaptive contrast."
    )
    grid_size: int = Field(
        default=8, ge=1, le=64, description="Grid size for adaptive contrast."
    )

    async def process(self, context: ProcessingContext) -> ImageRef:
        image = await context.image_to_pil(self.image)
        img = adaptive_contrast(
            image, clip_limit=self.clip_limit, grid_size=self.grid_size
        )
        return await context.image_from_pil(img)



#!/usr/bin/env python

from logging import Logger
from typing import IO, Any, AsyncIterator, Iterator

import httpx
from .abstract_storage import AbstractStorage

from nodetool.models.asset import Asset


class S3Storage(AbstractStorage):
    """
    This class, named `S3Storage`, is an implementation of the `AbstractStorage` class
    specifically designed to interact with Amazon S3 (Simple Storage Service) or
    compatible storage systems.

    The main purpose of this class is to provide methods for uploading, downloading,
    and deleting files (referred to as "objects" in S3 terminology) from an S3 bucket.
    It uses presigned URLs to perform these operations securely without exposing the
    actual S3 credentials.
    """

    bucket_name: str
    client: Any
    log: Logger
    endpoint_url: str | None = None
    domain: str | None = None

    def __init__(
        self,
        bucket_name: str,
        client: Any,
        log: Logger,
        endpoint_url: str | None = None,
        domain: str | None = None,
    ):
        self.bucket_name = bucket_name
        self.client = client
        self.log = log
        self.endpoint_url = endpoint_url
        self.domain = domain

    def get_base_url(self):
        """
        Get the base URL for the S3 bucket.
        """
        return f"{self.endpoint_url}/{self.bucket_name}"

    def get_url(self, key: str):
        """
        Get the URL for the given S3 object.
        """
        if self.domain:
            return f"https://{self.domain}/{key}"
        else:
            return f"{self.endpoint_url}/{self.bucket_name}/{key}"

    def generate_presigned_url(
        self,
        client_method: str,
        object_name: str,
        expiration=3600 * 24 * 7,
    ):
        """
        Generate a presigned URL for the given S3 object.
        """
        if self.endpoint_url and self.endpoint_url.startswith("http://localhost"):
            return f"{self.endpoint_url}/{self.bucket_name}/{object_name}"

        response = self.client.generate_presigned_url(
            client_method,
            Params={"Bucket": self.bucket_name, "Key": object_name},
            ExpiresIn=expiration,
        )

        return response

    async def file_exists(self, file_name: str) -> bool:
        """
        Check if an asset exists in S3.
        """
        url = self.generate_presigned_url("head_object", file_name)
        try:
            async with httpx.AsyncClient() as client:
                response = await client.head(url)
                response.raise_for_status()

            return True
        except:
            return False

    async def get_mtime(self, key: str):
        """
        Get the last modified time of the file.
        """
        url = self.generate_presigned_url("head_object", key)

        async with httpx.AsyncClient() as client:
            response = await client.head(url)
            response.raise_for_status()

            last_modified = response.headers.get("last-modified")
            return last_modified

    async def download(self, key: str, stream: IO):
        """
        Downloads a blob from the bucket.
        """
        url = self.get_url(key)
        async with httpx.AsyncClient() as client:
            async with client.stream("GET", url) as response:
                response.raise_for_status()

                async for chunk in response.aiter_bytes():
                    stream.write(chunk)

    async def upload(self, key: str, content: IO):
        """
        Uploads a blob to the bucket.
        """
        url = self.generate_presigned_url("put_object", key)

        async with httpx.AsyncClient() as client:
            response = await client.put(url, content=content.read())
            response.raise_for_status()

        return self.get_url(key)

    async def download_stream(self, key: str) -> AsyncIterator[bytes]:
        """
        Downloads a blob from the bucket as a stream.
        """
        url = self.get_url(key)

        async with httpx.AsyncClient() as client:
            async with client.stream("GET", url) as response:
                response.raise_for_status()
                async for chunk in response.aiter_bytes():
                    yield chunk

    async def delete(self, file_name: str):
        url = self.generate_presigned_url("delete_object", file_name)
        async with httpx.AsyncClient() as client:
            response = await client.delete(url)
            response.raise_for_status()



from typing import Any
from pydantic import Field
from nodetool.metadata.types import DataframeRef, ImageTensor
from nodetool.metadata.types import asset_to_ref
from nodetool.models.asset import Asset
from nodetool.metadata.types import FolderRef
from nodetool.metadata.types import AssetRef
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.metadata.types import AudioRef
from nodetool.metadata.types import ImageRef
from nodetool.metadata.types import TextRef
from nodetool.workflows.base_node import BaseNode, InputNode
from nodetool.metadata.types import VideoRef


class FloatInput(InputNode):
    """
    Represents a float parameter for the workflow.
    input, parameter, float, number
    """

    value: float = 0.0
    min: float = 0
    max: float = 100

    async def process(self, context: ProcessingContext) -> float:
        return min(max(self.value, self.min), self.max)

    def get_json_schema(self):
        return {
            "type": "number",
            "description": self.description,
            "default": self.value,
            "minimum": self.min,
            "maximum": self.max,
        }


class BooleanInput(InputNode):
    """
    Represents a boolean parameter for the workflow.
    input, parameter, boolean, bool
    """

    value: bool = False

    async def process(self, context: ProcessingContext) -> bool:
        return self.value

    def get_json_schema(self):
        return {
            "type": "boolean",
            "description": self.description,
            "default": self.value,
        }


class IntegerInput(InputNode):
    """
    Represents an integer parameter for the workflow.
    input, parameter, integer, number
    """

    value: int = 0
    min: int = 0
    max: int = 100

    async def process(self, context: ProcessingContext) -> int:
        return min(max(self.value, self.min), self.max)

    def get_json_schema(self):
        return {
            "type": "integer",
            "description": self.description,
            "default": self.value,
            "minimum": self.min,
            "maximum": self.max,
        }


class StringInput(InputNode):
    """
    Represents a string parameter for the workflow.
    input, parameter, string, text
    """

    value: str = ""

    async def process(self, context: ProcessingContext) -> str:
        return self.value

    def get_json_schema(self):
        return {
            "type": "string",
            "description": self.description,
            "default": self.value,
        }


class ChatInput(InputNode):
    """
    Represents a chat message parameter for the workflow.
    input, parameter, chat, message
    """

    value: str = ""

    async def process(self, context: ProcessingContext) -> str:
        return self.value

    def get_json_schema(self):
        return {
            "type": "string",
            "description": self.description,
            "default": self.value,
        }


class TextInput(InputNode):
    """
    Represents a text parameter for the workflow.
    input, parameter, text
    """

    value: TextRef = Field(TextRef(), description="The text to use as input.")

    async def process(self, context: ProcessingContext) -> TextRef:
        return self.value

    def get_json_schema(self):
        return {
            "type": "string",
            "description": self.description,
            "default": self.value,
        }


class AssetSchemaMixin:
    def get_json_schema(self):
        return {
            "type": "object",
            "properties": {
                "uri": {
                    "type": "string",
                    "description": "The URI of the image.",
                    "format": "uri",
                },
                "asset_id": {
                    "type": "string",
                    "description": "The Asset ID of the image.",
                },
            },
        }


class ImageInput(AssetSchemaMixin, InputNode):
    """
    Represents an image parameter for the workflow.
    input, parameter, image
    """

    value: ImageRef = Field(ImageRef(), description="The image to use as input.")

    async def process(self, context: ProcessingContext) -> ImageRef:
        return self.value


class VideoInput(AssetSchemaMixin, InputNode):
    """
    Represents a video parameter for the workflow.
    input, parameter, video
    """

    value: VideoRef = Field(VideoRef(), description="The video to use as input.")

    async def process(self, context: ProcessingContext) -> VideoRef:
        return self.value


class AudioInput(AssetSchemaMixin, InputNode):
    """
    Represents an audio parameter for the workflow.
    input, parameter, audio
    """

    value: AudioRef = Field(AudioRef(), description="The audio to use as input.")

    async def process(self, context: ProcessingContext) -> AudioRef:
        return self.value


class Folder(AssetSchemaMixin, InputNode):
    """
    Represents a folder parameter for the workflow.
    input, parameter, folder
    """

    folder: FolderRef = Field(FolderRef(), description="The folder to use as input.")
    limit: int = 1000

    async def process(self, context: ProcessingContext) -> list[AssetRef]:
        if self.folder.is_empty():
            return []

        assets, cursor = Asset.paginate(
            user_id=context.user_id,
            parent_id=self.folder.asset_id,
            limit=self.limit,
        )

        return [asset_to_ref(asset) for asset in assets]


class ImageFolder(Folder):
    """
    Represents an image folder parameter for the workflow.
    input, parameter, folder, image
    """

    async def process(self, context: ProcessingContext) -> list[ImageRef]:
        assets = await super().process(context)
        images = [asset for asset in assets if isinstance(asset, ImageRef)]
        return images


class AudioFolder(Folder):
    """
    Repesents an audio folder parameter for the workflow.
    input, parameter, folder, audio
    """

    async def process(self, context: ProcessingContext) -> list[AudioRef]:
        assets = await super().process(context)
        audios = [asset for asset in assets if isinstance(asset, AudioRef)]
        return audios


class VideoFolder(Folder):
    """
    Represents a video folder parameter for the workflow.
    input, parameter, folder, video
    """

    async def process(self, context: ProcessingContext) -> list[VideoRef]:
        assets = await super().process(context)
        videos = [asset for asset in assets if isinstance(asset, VideoRef)]
        return videos


class TextFolder(Folder):
    """
    Represents a text folder parameter for the workflow.
    input, parameter, folder, text
    """

    async def process(self, context: ProcessingContext) -> list[TextRef]:
        assets = await super().process(context)
        texts = [asset for asset in assets if isinstance(asset, TextRef)]
        return texts


class ComfyImageInput(AssetSchemaMixin, InputNode):
    """
    Represents an image parameter for the workflow.
    input, parameter, image
    """

    value: ImageRef = Field(ImageRef(), description="The image to use as input.")

    async def process(self, context: ProcessingContext) -> ImageTensor:
        from PIL import Image, ImageOps, ImageSequence
        import torch
        import numpy as np

        img = await context.image_to_pil(self.value)

        output_images = []
        output_masks = []
        for i in ImageSequence.Iterator(img):
            i = ImageOps.exif_transpose(i)
            image = i.convert("RGB")
            image = np.array(image).astype(np.float32) / 255.0
            image = torch.from_numpy(image)[None,]
            if "A" in i.getbands():
                mask = np.array(i.getchannel("A")).astype(np.float32) / 255.0
                mask = 1.0 - torch.from_numpy(mask)
            else:
                mask = torch.zeros((64, 64), dtype=torch.float32, device="cpu")
            output_images.append(image)
            output_masks.append(mask.unsqueeze(0))

        if len(output_images) > 1:
            output_image = torch.cat(output_images, dim=0)
            output_mask = torch.cat(output_masks, dim=0)
        else:
            output_image = output_images[0]
            output_mask = output_masks[0]

        return output_image  # type: ignore


class GroupInput(BaseNode):
    """
    Input node for any group node.
    """

    items: list[Any] | DataframeRef = []
    name: str = ""
    _value: Any = None

    async def process(self, context: Any) -> Any:
        return self._value



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class JoinImageWithAlpha(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to join with an alpha channel.')
    alpha: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The alpha channel (mask) to join with the image.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.compositing.JoinImageWithAlpha"


from nodetool.nodes.comfy.mask.compositing import PorterDuffModeEnum

class PorterDuffImageComposite(GraphNode):
    source: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The source image.')
    source_alpha: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The source alpha (mask).')
    destination: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The destination image.')
    destination_alpha: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The destination alpha (mask).')
    mode: PorterDuffModeEnum | GraphNode | tuple[GraphNode, str] = Field(default=PorterDuffModeEnum('DST'), description='The Porter-Duff compositing mode to use.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.compositing.PorterDuffImageComposite"



class SplitImageWithAlpha(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image with an alpha channel to split.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.compositing.SplitImageWithAlpha"





"""
Copyright (c) Microsoft Corporation.
Licensed under the MIT license.

Basic logger. It Computes and stores the average and current value
"""

class AverageMeter(object):
    
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count



class EvalMetricsLogger(object):
    
    def __init__(self):
        self.reset()

    def reset(self):
        # define a upper-bound performance (worst case) 
        # numbers are in unit millimeter
        self.PAmPJPE = 100.0/1000.0
        self.mPJPE = 100.0/1000.0
        self.mPVE = 100.0/1000.0

        self.epoch = 0

    def update(self, mPVE, mPJPE, PAmPJPE, epoch):
        self.PAmPJPE = PAmPJPE
        self.mPJPE = mPJPE
        self.mPVE = mPVE
        self.epoch = epoch



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class SetLatentNoiseMask(GraphNode):
    samples: Latent | GraphNode | tuple[GraphNode, str] = Field(default=Latent(type='comfy.latent'), description='The latent samples to set the noise mask for.')
    mask: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The mask to use for the noise.')
    @classmethod
    def get_node_type(cls): return "comfy.latent.inpaint.SetLatentNoiseMask"





'''
Copyright 2017 Javier Romero, Dimitrios Tzionas, Michael J Black and the Max Planck Gesellschaft.  All rights reserved.
This software is provided for research purposes only.
By using this software you agree to the terms of the MANO/SMPL+H Model license here http://mano.is.tue.mpg.de/license

More information about MANO/SMPL+H is available at http://mano.is.tue.mpg.de.
For comments or questions, please email us at: mano@tue.mpg.de


About this file:
================
This file defines a wrapper for the loading functions of the MANO model.

Modules included:
- load_model:
  loads the MANO model from a given file location (i.e. a .pkl file location),
  or a dictionary object.

'''


import numpy as np
import cv2

def lrotmin(p):
    if isinstance(p, np.ndarray):
        p = p.ravel()[3:]
        return np.concatenate(
            [(cv2.Rodrigues(np.array(pp))[0] - np.eye(3)).ravel()
             for pp in p.reshape((-1, 3))]).ravel()


def posemap(s):
    if s == 'lrotmin':
        return lrotmin
    else:
        raise Exception('Unknown posemapping: %s' % (str(s), ))


import functools
import os
import warnings

import cv2
import numpy as np
import torch
import torch.nn as nn
from einops import rearrange
from huggingface_hub import hf_hub_download
from PIL import Image

from controlnet_aux.util import HWC3, resize_image_with_pad, common_input_validate, custom_hf_download, HF_MODEL_NAME


class UnetGenerator(nn.Module):
    """Create a Unet-based generator"""

    def __init__(self, input_nc, output_nc, num_downs, ngf=64, norm_layer=nn.BatchNorm2d, use_dropout=False):
        """Construct a Unet generator
        Parameters:
            input_nc (int)  -- the number of channels in input images
            output_nc (int) -- the number of channels in output images
            num_downs (int) -- the number of downsamplings in UNet. For example, # if |num_downs| == 7,
                                image of size 128x128 will become of size 1x1 # at the bottleneck
            ngf (int)       -- the number of filters in the last conv layer
            norm_layer      -- normalization layer
        We construct the U-Net from the innermost layer to the outermost layer.
        It is a recursive process.
        """
        super(UnetGenerator, self).__init__()
        # construct unet structure
        unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, input_nc=None, submodule=None, norm_layer=norm_layer, innermost=True)  # add the innermost layer
        for _ in range(num_downs - 5):          # add intermediate layers with ngf * 8 filters
            unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, input_nc=None, submodule=unet_block, norm_layer=norm_layer, use_dropout=use_dropout)
        # gradually reduce the number of filters from ngf * 8 to ngf
        unet_block = UnetSkipConnectionBlock(ngf * 4, ngf * 8, input_nc=None, submodule=unet_block, norm_layer=norm_layer)
        unet_block = UnetSkipConnectionBlock(ngf * 2, ngf * 4, input_nc=None, submodule=unet_block, norm_layer=norm_layer)
        unet_block = UnetSkipConnectionBlock(ngf, ngf * 2, input_nc=None, submodule=unet_block, norm_layer=norm_layer)
        self.model = UnetSkipConnectionBlock(output_nc, ngf, input_nc=input_nc, submodule=unet_block, outermost=True, norm_layer=norm_layer)  # add the outermost layer

    def forward(self, input):
        """Standard forward"""
        return self.model(input)


class UnetSkipConnectionBlock(nn.Module):
    """Defines the Unet submodule with skip connection.
        X -------------------identity----------------------
        |-- downsampling -- |submodule| -- upsampling --|
    """

    def __init__(self, outer_nc, inner_nc, input_nc=None,
                 submodule=None, outermost=False, innermost=False, norm_layer=nn.BatchNorm2d, use_dropout=False):
        """Construct a Unet submodule with skip connections.
        Parameters:
            outer_nc (int) -- the number of filters in the outer conv layer
            inner_nc (int) -- the number of filters in the inner conv layer
            input_nc (int) -- the number of channels in input images/features
            submodule (UnetSkipConnectionBlock) -- previously defined submodules
            outermost (bool)    -- if this module is the outermost module
            innermost (bool)    -- if this module is the innermost module
            norm_layer          -- normalization layer
            use_dropout (bool)  -- if use dropout layers.
        """
        super(UnetSkipConnectionBlock, self).__init__()
        self.outermost = outermost
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm2d
        else:
            use_bias = norm_layer == nn.InstanceNorm2d
        if input_nc is None:
            input_nc = outer_nc
        downconv = nn.Conv2d(input_nc, inner_nc, kernel_size=4,
                             stride=2, padding=1, bias=use_bias)
        downrelu = nn.LeakyReLU(0.2, True)
        downnorm = norm_layer(inner_nc)
        uprelu = nn.ReLU(True)
        upnorm = norm_layer(outer_nc)

        if outermost:
            upconv = nn.ConvTranspose2d(inner_nc * 2, outer_nc,
                                        kernel_size=4, stride=2,
                                        padding=1)
            down = [downconv]
            up = [uprelu, upconv, nn.Tanh()]
            model = down + [submodule] + up
        elif innermost:
            upconv = nn.ConvTranspose2d(inner_nc, outer_nc,
                                        kernel_size=4, stride=2,
                                        padding=1, bias=use_bias)
            down = [downrelu, downconv]
            up = [uprelu, upconv, upnorm]
            model = down + up
        else:
            upconv = nn.ConvTranspose2d(inner_nc * 2, outer_nc,
                                        kernel_size=4, stride=2,
                                        padding=1, bias=use_bias)
            down = [downrelu, downconv, downnorm]
            up = [uprelu, upconv, upnorm]

            if use_dropout:
                model = down + [submodule] + up + [nn.Dropout(0.5)]
            else:
                model = down + [submodule] + up

        self.model = nn.Sequential(*model)

    def forward(self, x):
        if self.outermost:
            return self.model(x)
        else:   # add skip connections
            return torch.cat([x, self.model(x)], 1)


class LineartAnimeDetector:
    def __init__(self, model):
        self.model = model

    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, filename="netG.pth"):
        model_path = custom_hf_download(pretrained_model_or_path, filename)

        norm_layer = functools.partial(nn.InstanceNorm2d, affine=False, track_running_stats=False)
        net = UnetGenerator(3, 1, 8, 64, norm_layer=norm_layer, use_dropout=False)
        ckpt = torch.load(model_path)
        for key in list(ckpt.keys()):
            if 'module.' in key:
                ckpt[key.replace('module.', '')] = ckpt[key]
                del ckpt[key]
        net.load_state_dict(ckpt)
        net.eval()

        return cls(net)

    def to(self, device):
        self.model.to(device)
        return self
    
    def __call__(self, input_image, detect_resolution=512, output_type="pil", upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)

        H, W, C = input_image.shape
        Hn = 256 * int(np.ceil(float(H) / 256.0))
        Wn = 256 * int(np.ceil(float(W) / 256.0))
        input_image = cv2.resize(input_image, (Wn, Hn), interpolation=cv2.INTER_CUBIC)

        device = next(iter(self.model.parameters())).device
        with torch.no_grad():
            image_feed = torch.from_numpy(input_image).float().to(device)
            image_feed = image_feed / 127.5 - 1.0
            image_feed = rearrange(image_feed, 'h w c -> 1 c h w')

            line = self.model(image_feed)[0, 0] * 127.5 + 127.5
            line = line.cpu().numpy()
            line = line.clip(0, 255).astype(np.uint8)
        
        #A1111 uses INTER AREA for downscaling so ig that is the best choice
        detected_map = cv2.resize(HWC3(line), (W, H), interpolation=cv2.INTER_AREA)
        detected_map = remove_pad(255 - detected_map)

        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map



from queue import Queue
from nodetool.common.environment import Environment
from nodetool.workflows.run_job_request import RunJobRequest

import asyncio
import threading
import time
from sys import stdout

log = Environment.get_logger()


def run_workflow(req: RunJobRequest):
    from nodetool.workflows.types import Error, WorkflowUpdate
    from nodetool.workflows.processing_context import (
        ProcessingContext,
    )
    from nodetool.workflows.workflow_runner import WorkflowRunner

    context = ProcessingContext(
        user_id=req.user_id,
        auth_token=req.auth_token,
        workflow_id=req.workflow_id,
        queue=Queue(),
    )

    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    runner = WorkflowRunner()

    async def run():
        try:
            await runner.run(req, context)
        except Exception as e:
            log.exception(e)
            context.post_message(Error(error=str(e)))

    thread = threading.Thread(target=lambda: asyncio.run(run()))
    thread.start()

    try:
        while runner.is_running():
            if context.has_messages():
                msg = context.pop_message()
                yield msg.model_dump_json() + "\n"
                if isinstance(msg, Error):
                    break
                elif isinstance(msg, WorkflowUpdate):
                    pass
            else:
                time.sleep(0.1)

        while context.has_messages():
            msg = context.pop_message()
            yield msg.model_dump_json() + "\n"

        thread.join()

    except Exception as e:
        log.exception(e)
        yield Error(error=str(e)).model_dump_json() + "\n"



from pydantic import Field
from nodetool.metadata.types import Latent, Mask
from nodetool.common.comfy_node import ComfyNode


class SetLatentNoiseMask(ComfyNode):
    """
    The Set Latent Noise Mask node can be used to add a mask to the latent images for inpainting. When the noise mask is set a sampler node will only operate on the masked area. If a single mask is provided, all the latents in the batch will use this mask.
    """

    samples: Latent = Field(
        default=Latent(), description="The latent samples to set the noise mask for."
    )
    mask: Mask = Field(default=Mask(), description="The mask to use for the noise.")

    @classmethod
    def return_type(cls):
        return {"latent": Latent}



from enum import Enum
from pydantic import Field
from nodetool.metadata.types import AudioRef
from nodetool.metadata.types import ImageRef
from nodetool.common.huggingface_node import HuggingfaceNode
from nodetool.workflows.processing_context import ProcessingContext


class ModelId(str, Enum):
    OPENAI_WHISPER_LARGE_V3 = "openai/whisper-large-v3"
    OPENAI_WHISPER_LARGE_V2 = "openai/whisper-large-v2"


class AutomaticSpeechRecognition(HuggingfaceNode):
    """
    Automatic Speech Recognition (ASR), also known as Speech to Text (STT), is the task of transcribing a given audio to text. It has many applications, such as voice user interfaces.
    asr, speech, audio, huggingface
    """

    model: ModelId = Field(
        default=ModelId.OPENAI_WHISPER_LARGE_V3,
        title="Model ID on Huggingface",
        description="The model ID to use for the speech recognition",
    )
    audio: AudioRef = Field(
        default=AudioRef(),
        title="Image",
        description="The input audio to transcribe",
    )

    async def process(self, context: ProcessingContext) -> str:
        audio = await context.asset_to_io(self.audio)
        result = await self.run_huggingface(
            model_id=self.model, context=context, data=audio.read()
        )
        return result["text"]  # type: ignore



from typing import Any, List
from nodetool.api.types.graph import Graph
from nodetool.common.environment import Environment
from nodetool.models.asset import Asset as AssetModel
from nodetool.models.workflow import Workflow as WorkflowModel

from pydantic import BaseModel


class Workflow(BaseModel):
    id: str
    access: str
    created_at: str
    updated_at: str
    name: str
    description: str
    thumbnail: str | None = None
    thumbnail_url: str | None = None
    graph: Graph
    input_schema: dict[str, Any] | None = None
    output_schema: dict[str, Any] | None = None

    @classmethod
    def from_model(cls, workflow: WorkflowModel):
        thumbnail_url = None
        if workflow.thumbnail and workflow.thumbnail != "":
            asset = AssetModel.get(workflow.thumbnail)
            if asset:
                thumbnail_url = Environment.get_asset_storage().get_url(asset.file_name)

        graph = workflow.get_graph()
        api_graph = workflow.get_api_graph()

        return cls(
            id=workflow.id,
            access=workflow.access,
            created_at=workflow.created_at.isoformat(),
            updated_at=workflow.updated_at.isoformat(),
            name=workflow.name,
            description=workflow.description or "",
            thumbnail=workflow.thumbnail or "",
            thumbnail_url=thumbnail_url,
            graph=api_graph,
            input_schema=graph.get_input_schema(),
            output_schema=graph.get_output_schema(),
        )


class WorkflowRequest(BaseModel):
    name: str
    description: str
    thumbnail: str | None = None
    access: str
    graph: Graph | None = None
    comfy_workflow: dict[str, Any] | None = None


class WorkflowList(BaseModel):
    next: str | None
    workflows: List[Workflow]
